# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.deep_merger import DeepMerger, ListTuple


class ExplodeStructureTests(unittest.TestCase):
    def test_merge_appends_flat_dicts_as_items_in_list(self):
        # [{"dict_1": "value_1"}, {"dict_2": "value_2"}, {"dict_3": "value_3"}]
        base = [{"list_one": [ListTuple({"dict_1": "value_1"}, 0), ListTuple({"dict_2": "value_2"}, 1)]}]
        nxt = {'list_one': [ListTuple(value={'dict_3': 'value_3'}, position=2)]}
        self.assertEqual({"list_one": [ListTuple({"dict_1": "value_1"}, 0),
                                       ListTuple({"dict_2": "value_2"}, 1),
                                       ListTuple({'dict_3': 'value_3'}, 2)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_nested_list_within_list(self):
        # ["item1", "item2"]
        base = [{"lvl1": [ListTuple("item1", 0)]}]
        nxt = {'lvl1': [ListTuple("item_2", 1)]}
        self.assertEqual({'lvl1': [ListTuple("item1", 0), ListTuple("item_2", 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_deeply_nested_list_within_list(self):
        # ['item_1', [[['item_2']]]]
        base = [{"lvl1": [ListTuple("item1", 0)]}]
        nxt = {'lvl1': [ListTuple([ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 0)], 1)]}
        self.assertEqual({'lvl1': [ListTuple("item1", 0), ListTuple([ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 0)], 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_after_nested_list(self):
        # ['item_1', [[['item_2']]], "item_3"]
        base = [{'lvl1': [ListTuple('item_1', 0), ListTuple([ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 0)], 1)]}]
        self.assertEqual({'lvl1': [ListTuple('item_1', 0), ListTuple([ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 0)], 1), ListTuple("item_3", 2)]},
                         DeepMerger.merge(base, {'lvl1': [ListTuple("item_3", 2)]}))

    def test_merger_appends_within_nested_list(self):
        # ['item_1', ['item_2', "item_3"]]
        base = [{'lvl1': [ListTuple("item_1", 0), ListTuple([ListTuple('item_2', 0)], 1)]}]
        nxt = {'lvl1': [ListTuple([ListTuple("item_3", 1)], 1)]}
        self.assertEqual({'lvl1': [ListTuple('item_1', 0), ListTuple([ListTuple('item_2', 0), ListTuple('item_3', 1)], 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_merges_dicts_with_a_single_key_value_pair(self):
        # [item_1, {"lvl2" : "item_2"}, {"lvl3" : "item_3"}]
        base = [{'lvl1': [ListTuple('item_1', 0), ListTuple({"lvl2": "item_2"}, 1)]}]
        nxt = {'lvl1': [ListTuple({"lvl3": "item_3"}, 2)]}
        self.assertEqual({'lvl1': [ListTuple('item_1', 0), ListTuple({"lvl2": "item_2"}, 1), ListTuple({"lvl3": "item_3"}, 2)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_merges_dicts_with_multiple_key_value_pairs(self):
        # {lvl1: [item_1, {"lvl2" : "item_2", "lvl3" : "item_3"}]}
        base = [{'lvl1': [ListTuple('item_1', 0), ListTuple({"lvl2": "item_2"}, 1)]}]
        nxt = {'lvl1': [ListTuple({"lvl3": "item_3"}, 1)]}
        self.assertEqual({'lvl1': [ListTuple('item_1', 0), ListTuple({"lvl2": "item_2", "lvl3": "item_3"}, 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_merges_lists_at_depth(self):
        # ["item_0", ["item_1", [["item_2"]]]]
        base = [{"nested_list_1": [ListTuple("item_0", 0), ListTuple([ListTuple("item_1", 0)], 1)]}]
        nxt = {"nested_list_1": [ListTuple([ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 1)], 1)]}
        self.assertEqual({"nested_list_1": [ListTuple("item_0", 0),
                                            ListTuple([ListTuple("item_1", 0), ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 1)], 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_dict_at_correct_depth_when_previous_element_is_deeper(self):
        # {'nested_list_1': ['item_0', ['item_1', [['item_2']], {'item_3': {'lvl1': 'value_5'}}]]}
        base = [{'nested_list_1': [ListTuple("item_0", 0),
                                   ListTuple([ListTuple("item_1", 0), ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 1)], 1)]}]
        nxt = {'nested_list_1': [ListTuple([ListTuple({'item_3': {'lvl1': 'value_5'}}, 2)], 1)]}
        self.assertEqual({'nested_list_1': [ListTuple("item_0", 0),
                                            ListTuple([ListTuple("item_1", 0),
                                                       ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 1),
                                                       ListTuple({'item_3': {'lvl1': 'value_5'}}, 2)], 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_merges_dicts_with_lists(self):
        # {"lvl1": {"lvl2": {"lvl3_1": "value_1", "lvl3_2": [["value_2"]]
        base = [{'lvl1': {'lvl2': {'lvl3_1': 'value_1'}}}]
        nxt = {'lvl1': {'lvl2': {'lvl3_2': [ListTuple([ListTuple('value_2', 0)], 0)]}}}
        self.assertEqual({"lvl1": {"lvl2": {"lvl3_1": "value_1", "lvl3_2": [ListTuple([ListTuple('value_2', 0)], 0)]}}},
                         DeepMerger.merge(base, nxt))

    def test_merger_merges_dicts_with_nested_lists(self):
        # {"lvl1": {"lvl2": {"lvl3_1": "value_1", "lvl3_2": ["value_2", ['value_3]]}
        base = [{'lvl1': {'lvl2': {'lvl3_1': 'value_1', 'lvl3_2': [ListTuple("value_2", 0)]}}}]
        nxt = {'lvl1': {'lvl2': {'lvl3_2': [ListTuple([ListTuple('value_3', 0)], 1)]}}}
        self.assertEqual({"lvl1": {"lvl2": {"lvl3_1": 'value_1', "lvl3_2": [ListTuple("value_2", 0), ListTuple([ListTuple('value_3', 0)], 1)]}}},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_list_to_dict(self):
        # {"lvl1": {"lvl2": {"lvl3_1": {'key_1': ['value_1']},
        #                    "lvl3_2": {"lvl4_1": "value_2"},
        #                    "lvl3_3": [['value_3']]}}}

        base = [{'lvl1': {'lvl2': {'lvl3_1': {'key_1': [ListTuple('value_1', 0)]}, 'lvl3_2': {"lvl4_1": "value_2"}}}}]
        nxt = {'lvl1': {'lvl2': {'lvl3_3': [ListTuple([ListTuple('value_3', 0)], 0)]}}}
        self.assertEqual({"lvl1": {"lvl2": {"lvl3_1": {'key_1': [ListTuple('value_1', 0)]}, "lvl3_2": {"lvl4_1": "value_2"},
                                            "lvl3_3": [ListTuple([ListTuple('value_3', 0)], 0)]}}},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_list_to_dict_when_previous_element_is_list(self):
        # {"lvl1": {"lvl2": {"lvl3_1": ['value_1'],
        #                    "lvl3_2": [['value_2']]}}}

        base = [{'lvl1': {'lvl2': {'lvl3_1': [ListTuple('value_1', 0)]}}}]
        nxt = {'lvl1': {'lvl2': {'lvl3_2': [ListTuple([ListTuple('value_2', 0)], 0)]}}}
        self.assertEqual({"lvl1": {"lvl2": {"lvl3_1": [ListTuple('value_1', 0)], "lvl3_2": [ListTuple([ListTuple('value_2', 0)], 0)]}}},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_to_an_entire_list_at_depth(self):
        # {"nested_list_1": ["item1", 50, True, ["item2_1", "item2_2", {"item2_3": "value_1"}]]}
        base = [{"nested_list_1": [ListTuple("item1", 0), ListTuple(50, 1), ListTuple(True, 2), ListTuple([ListTuple("item2_1", 0), ListTuple("item2_2", 1)], 3)]}]
        nxt = {'nested_list_1': [ListTuple([ListTuple({'item2_3': 'value_1'}, 2)], 3)]}
        self.assertEqual({"nested_list_1": [ListTuple("item1", 0), ListTuple(50, 1), ListTuple(True, 2), ListTuple([ListTuple("item2_1", 0), ListTuple("item2_2", 1), ListTuple({'item2_3': 'value_1'}, 2)], 3)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_to_multiple_nested_list(self):
        # ["item_0", ["item_1", [["item_2", "item_3]]]]
        base = [{"nested_list_1": [ListTuple("item_0", 0),
                                   ListTuple([ListTuple("item_1", 0), ListTuple([ListTuple([ListTuple("item_2", 0)], 0)], 1)], 1)]}]
        nxt = {"nested_list_1": [ListTuple([ListTuple([ListTuple([ListTuple("item_3", 1)], 0)], 1)], 1)]}

        self.assertEqual({"nested_list_1": [ListTuple("item_0", 0),
                                            ListTuple([ListTuple("item_1", 0), ListTuple([ListTuple([ListTuple("item_2", 0), ListTuple("item_3", 1)], 0)], 1)], 1)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_dicts_in_nested_dicts_inside_a_list_item(self):
        # {'list1': [{'item_0': {'lvl1': {'lvl2_1': {'lvl3_1': 'value_1'},
        #                                 'lvl2_2': {'lvl3_2': 'value_2'}}}}]}

        base = [{'list1': [ListTuple({'item_0': {'lvl1': {'lvl2_1': {'lvl3_1': 'value_1'}}}}, 0)]}]
        nxt = {'list1': [ListTuple({'item_0': {'lvl1': {'lvl2_2': {'lvl3_2': 'value_2'}}}}, 0)]}
        self.assertEqual({'list1': [ListTuple({'item_0': {'lvl1': {'lvl2_1': {'lvl3_1': 'value_1'},
                                                         'lvl2_2': {'lvl3_2': 'value_2'}}}}, 0)]},
                         DeepMerger.merge(base, nxt))

    def test_merger_appends_dicts_in_dicts_inside_lists(self):
        # {'list1': [{'lvl1': {'list2': [{'item1': {'key1': 'value_1'},
        #                                 'item2': {'key2': 'value_2'}}]}}]}
        base = [{'list1': [ListTuple({'lvl1': {'list2': [ListTuple({'item1': {'key1': 'value_1'}}, 0)]}}, 0)]}]
        nxt = {'list1': [ListTuple({'lvl1': {'list2': [ListTuple({'item2': {'key2': 'value_2'}}, 0)]}}, 0)]}

        self.assertEqual({'list1': [ListTuple({'lvl1': {'list2': [ListTuple({'item1': {'key1': 'value_1'},
                                                                             'item2': {'key2': 'value_2'}}, 0)]}}, 0)]},
                         DeepMerger.merge(base, nxt))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

from pysisl.deep_merger import DeepMerger, ListTuple


class ExplodeStructure:
    @classmethod
    def explode_structure(cls, structure):
        """
        Expands input structure to list, from first item adding successive items as new lists
        """
        return cls._merge_paths(cls.enumerate_paths(structure))

    @classmethod
    def _merge_paths(cls, leaves):
        output_list = [[next(leaves)][0]]
        for each_element in leaves:
            output_list.append(DeepMerger.merge(output_list, each_element))
        return output_list

    @classmethod
    def enumerate_paths(cls, p):
        if type(p) is list:
            yield from cls._enumerate_list(p)
        elif type(p) is dict:
            yield from cls._enumerate_dict(p)
        elif type(p) is ListTuple:
            yield from cls._enumerate_list_tuple(p)
        else:
            yield p

    @classmethod
    def _enumerate_list(cls, p):
        """
        Expands list to nested list, from first element adding successive elements as ListTuples
        [1, 2, 3]
        -> [[ListTuple(1, 0)], [ListTuple(1, 0), ListTuple(2, 1)], [ListTuple(1, 0), ListTuple(2, 1), ListTuple(3, 2)]]
        """
        for index, list_item in enumerate(p):
            for item in cls.enumerate_paths(list_item):
                if type(item) == ListTuple:
                    yield [item]
                else:
                    yield [ListTuple(item, index)]

    @classmethod
    def _enumerate_dict(cls, p):
        """
        Expands dictionary to list, from first {key, value} pair adding successive {key, value} pairs
        {"key": 1, "key2": 2, "key3": 3}
        -> [{key1: 1}, {"key": 1, "key2": 2}, {"key": 1, "key2": 2, "key3": 3}]
        """
        for key, value in p.items():
            for path in cls.enumerate_paths(value):
                yield {key: path}

    @classmethod
    def _enumerate_list_tuple(cls, p):
        """
        Expands ListTuples to list, from first element adding successive elements
        [ListTuple(1, 0), ListTuple(2, 1), ListTuple(3, 1)]
        -> [[ListTuple(1, 0)], [ListTuple(1, 0), ListTuple(2, 1)], [ListTuple(1, 0), ListTuple(2, 1), ListTuple(3, 2)]]
        """
        if isinstance(p.value, (dict, list)):
            for value in cls.enumerate_paths(p.value):
                yield ListTuple(value, p.position)
        else:
            yield p




# Copyright PA Knowledge Ltd 2021

from .sisl_lexer import *
from .sisl_parser import *
from .sisl_joiner import *

import ply.lex as lex
import ply.yacc as yacc

from . import sisl_type_flatten

from jsonschema import validate as json_validator
from jsonschema import FormatChecker, ValidationError


class SislDecoder:
    def __init__(self):
        self.parser = yacc.yacc()
        lex.lex()

    def loads(self, sisl, schema=None):
        if type(sisl) is list:
            sisl = JoinSisl.join([self.parse_raw_types(parsed_sisl) for parsed_sisl in sisl])
        else:
            sisl = self.parse_raw_types(sisl)

        flattened_sisl = sisl_type_flatten.SISLTypeFlatten().flatten(sisl)
        self._verify_schema_if_required(flattened_sisl, schema)
        return flattened_sisl

    @staticmethod
    def _verify_schema_if_required(flattened_sisl, schema):
        if schema is None:
            return

        try:
            json_validator(flattened_sisl, schema=schema, format_checker=FormatChecker())
        except ValidationError as err:
            raise SislValidationError(err)

    def parse_raw_types(self, sisl):
        return self.parser.parse(sisl)


class SislValidationError(Exception):
    pass



# Copyright PA Knowledge Ltd 2021

#
# rYAMLfile       = grouping *255WSP
#
# grouping        = "{" *255WSP ( name ":" 1*255WSP "!" type 1*255WSP value *255WSP *( "," *255WSP name ":" 1*255WSP "!" type 1*255WSP value *255WSP ) / "" ) "}"
#
# name            = ( "_" / ALPHA ) *( "_" / "-" / "." / ALPHA / DIGIT )
#
# type            = ( "_" / ALPHA ) *254( "_" / "-" / "." / ALPHA / DIGIT )
#
# value           = ( DQUOTE *( PRINTABLE / ESCAPE) DQUOTE ) / grouping
#
# ESCAPE          = "\" ( LCR / LCT / LCN / DQUOTE / "\" / (LCX 2HEX) / (LCU 4HEX) / (UCU 8HEX) )
#
# HEX             = DIGIT / %x41-46 / %x61-66
#
# WSP             = SP / HTAB / CR / LF
#
# PRINTABLE       = %x20-21 / %x23-5B / %x5D-7E             ; Printable chars apart from '"' or '\'
# ALPHA           = %x41-5A / %x61-7A                       ; A-Z / a-z
# DIGIT           = %x30-39                                 ; 0-9
# DQUOTE          = %x22                                    ; " (double-quote)
# SP              = %x20                                    ; space
# HTAB            = %x09                                    ; horizontal tab
# CR              = %x0D                                    ; carriage return
# LF              = %x0A                                    ; line feed
# LCR             = %x72                                    ; lower case r
# LCT             = %x74                                    ; lower case t
# LCN             = %x6E                                    ; lower case n
# LCX             = %x78                                    ; lower case x
# LCU             = %x75                                    ; lower case u
# UCU             = %x55                                    ; upper case u
from .parser_error import ParserError


tokens = (
    'BEGIN_OBJECT',
    'END_OBJECT',
    'NAME',
    'TYPE',
    'VALUE',
    'LIST_SEPARATOR'
)

t_BEGIN_OBJECT = '{[\s]{0,255}'
t_END_OBJECT = '[\s]{0,255}}[\s]{0,255}'
t_LIST_SEPARATOR = r'[\s]{0,255},[\s]{0,255}'


# name = ( "_" / ALPHA ) *( "_" / "-" / "." / ALPHA / DIGIT ):
def t_NAME(t):
    r'[_a-zA-Z][a-zA-Z0-9-\_\.]*:[\s]{1,255}'
    t.value = t.value.rstrip()[:-1]
    return t


# type = ( "_" / ALPHA ) *254( "_" / "-" / "." / ALPHA / DIGIT )
def t_TYPE(t):
    r'\![_a-zA-Z][a-zA-Z0-9-\_\.]{0,254}[\s]{1,255}'
    t.value = t.value.rstrip()[1:]
    return t


# value = ( DQUOTE *( PRINTABLE / ESCAPE) DQUOTE ) / grouping
# ESCAPE = "\" ( LCR / LCT / LCN / DQUOTE / "\" / (LCX 2HEX) / (LCU 4HEX) / (UCU 8HEX) )
def t_VALUE(t):
    r'"([^"\\]|\\.)*"(?=\s{0,255},|\s{0,255}})'
    t.value = t.value[1:-1]  # Remove quotation marks
    return t


def t_error(t):
    print(f"Illegal character '{t.value[0]}'")
    raise ParserError(t)



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_encoder import SislEncoder
from pysisl.sisl_decoder import SislDecoder
import subprocess
import os


class RandomABNFTests(unittest.TestCase):
    def test_basic_sisl_parsed_through_pysisl(self):
        for abnf_generated_file in os.listdir("test_files/"):
            with open(f"test_files/{abnf_generated_file}", "r") as input_sisl:
                test_sisl = r'{}'.format((input_sisl.read()))
            converted = SislEncoder.dumps(SislDecoder().parse_raw_types(test_sisl))
            with open(f"test_output_files/{abnf_generated_file}", "w") as output_sisl:
                output_sisl.write(converted)

            result = subprocess.check_output(["./grammar.exe", "-d", "0", f"test_output_files/{abnf_generated_file}"],
                                             stderr=subprocess.STDOUT)
            if 'PASS' not in str(result):
                print(f"Failed on file:\n{abnf_generated_file}\nContents:\n{test_sisl[:20]}...")
                print(f"Error Message:\n{result}")
            self.assertTrue('PASS' in str(result))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

import unittest

from pysisl.sisl_splitter import SplitSisl


class SislSplitterTests(unittest.TestCase):
    def test_basic_dict_is_parsed_without_splitting(self):
        self.assertEqual(['{abc: !int "2"}'], SplitSisl(20).split_sisl({"abc": 2}))

    def test_basic_anonymous_string_is_parsed_without_splitting(self):
        self.assertEqual(['{_: !_str "abc"}'], SplitSisl(20).split_sisl("abc"))

    def test_basic_anonymous_bool_is_parsed_without_splitting(self):
        self.assertEqual(['{_: !_bool "true"}'], SplitSisl(20).split_sisl(True))

    def test_basic_anonymous_int_is_parsed_without_splitting(self):
        self.assertEqual(['{_: !_int "1"}'], SplitSisl(20).split_sisl(1))

    def test_basic_anonymous_nonetype_is_parsed_without_splitting(self):
        self.assertEqual(['{_: !_null ""}'], SplitSisl(20).split_sisl(None))

    def test_basic_anonymous_list_is_split(self):
        self.assertEqual(["{_: !_list {_0: !str \"testing this long \"}}", "{_: !_list {_1: !str \"string still fits\"}}"], SplitSisl(50).split_sisl(["testing this long ", "string still fits"]))

    def test_anonymous_list_with_dict_is_split(self):
        self.assertEqual(["{_: !_list {_0: !obj {field_one: !int \"3\"}}}", "{_: !_list {_0: !obj {field_two: !int \"2\"}}}"], SplitSisl(50).split_sisl([{"field_one": 3, "field_two": 2}]))

    def test_simple_sisl_string_longer_than_max_length_returns_list_of_two_sisl_strings_of_valid_length(self):
        self.assertEqual(["{field_one: !str \"string\"}","{field_two: !str \"string2\"}"],
                         SplitSisl(30).split_sisl({"field_one": "string", "field_two": "string2"}))

    def test_splitter_splits_to_fit_long_second_nested_object(self):
        self.assertEqual(["{field_one: !obj {field_two: !str \"hello\"}}", "{field_one: !obj {field_three: !str \"superlooooooooooooooooooooooooooooooooong\"}}"],
                         SplitSisl(81).split_sisl({"field_one": {"field_two": "hello", "field_three": "superlooooooooooooooooooooooooooooooooong"}}))

    def test_splitter_raises_an_error_if_object_cannot_be_parsed(self):
        with self.subTest("splitter_raises_an_error_if_length_is_set_too_low"):
            self.assertRaises(Exception, SplitSisl(41).split_sisl, {"field_one": {"field_two": "hello"}})
        with self.subTest("splitter_raises_an_error_if_second_nested_object_is_too_long"):
            self.assertRaises(Exception, SplitSisl(60).split_sisl, {"field_one": {"field_two": "hello", "field_three": "superlooooooooooooooooooooooooooooooooong"}})

    def test_nested_dict_is_parsed_without_splitting(self):  # p5
        self.assertEqual(['{abc: !int "2", def: !obj {efw: !str "wer", ggg: !int "234"}, adsfs: !str "hsadfklsdjlskdfjkl"}', '{ykies: !obj {whoami: !str "user", oooh: !obj {another: !float "123.2"}}}'],
                         SplitSisl(95).split_sisl({"abc": 2, "def": {"efw":"wer", "ggg": 234}, "adsfs": "hsadfklsdjlskdfjkl", "ykies": {"whoami": "user", "oooh": {"another": 123.2}}}))

    def test_nested_dict_split_up(self):
        self.assertEqual(['{field_one: !obj {field_two: !int "123"}}', '{field_one: !obj {field_three: !obj {field_four: !str "field_four string"}}}'],
                         SplitSisl(95).split_sisl({"field_one": {"field_two": 123, "field_three": {"field_four": "field_four string"}}}))

    def test_super_nested_items_can_be_fitted(self):
        self.assertEqual(["{field_one: !obj {field_two: !obj {field_three: !obj {field_four: !str \"value_1\"}}}}",
                         "{field_one: !obj {field_two: !obj {field_three: !obj {field_five: !obj {field_six: !obj {field_seven: !obj {field_eight: !str \"value_2\"}}}}}}}",
                         "{field_one: !obj {field_two: !obj {field_three: !obj {field_five: !obj {field_six: !obj {field_seven: !obj {field_nine: !str \"value_3\"}}}}}}}",
                         "{field_one: !obj {field_two: !obj {field_ten: !str \"value_4\"}}}"],
                         SplitSisl(142).split_sisl({"field_one": {"field_two": {"field_three": {"field_four": "value_1",
                                                                                               "field_five": {"field_six": {"field_seven": {"field_eight": "value_2", "field_nine": "value_3"}}}},
                                                                               "field_ten": "value_4"}}}))

    def test_multi_nested_and_long_value_dict_is_split(self):
        test_json = {"nest_one": {"sub_nest_one": {"nested_field_one": "test 1", "nested_field_two": "testing this long string still fits"}}}
        self.assertEqual(["{nest_one: !obj {sub_nest_one: !obj {nested_field_one: !str \"test 1\"}}}",
                         "{nest_one: !obj {sub_nest_one: !obj {nested_field_two: !str \"testing this long string still fits\"}}}"],
                         SplitSisl(100).split_sisl(test_json))

    def test_simple_list_is_handled(self):
        self.assertEqual(["{field_one: !list {_0: !str \"I\", _1: !str \"am\", _2: !str \"a\", _3: !str \"list\"}}"],
                         SplitSisl().split_sisl({"field_one": ["I", "am", "a", "list"]}))

    def test_simple_list_is_split_in_half(self):
        self.assertEqual(["{field_one: !list {_0: !str \"I\", _1: !str \"am\"}}", "{field_one: !list {_2: !str \"a\", _3: !str \"list\"}}"],
                         SplitSisl(50).split_sisl({"field_one": ["I", "am", "a", "list"]}))

    def test_simple_nested_list_is_split_in_half(self):
        self.assertEqual(["{field_one: !list {_0: !str \"I\", _1: !list {_0: !str \"am\"}}}", "{field_one: !list {_1: !list {_1: !str \"a\"}, _2: !str \"list\"}}"],
                         SplitSisl(65).split_sisl({"field_one": ["I", ["am", "a"], "list"]}))

    def test_long_list_item_is_split(self):
        self.assertEqual(["{field_one: !list {_0: !str \"I\", _1: !str \"am\", _2: !str \"a\"}}",
                         "{field_one: !list {_3: !str \"supeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeerlist\"}}"],
                         SplitSisl(120).split_sisl({"field_one": ["I", "am", "a", "supeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeerlist"]}))

    def test_splitter_splits_complex_and_multiple_nesting(self):
        test_json = {"nest_one": {"field_one": "hello",
                                  "field_two_nest": {"nested_field_one": "brave",
                                                     "nested_field_two": "new"},
                                  "field_three": "world"},
                     "nest_two": {"field_one": "world",
                                  "field_two": "wide"}}
        self.assertEqual(["{nest_one: !obj {field_one: !str \"hello\", field_two_nest: !obj {nested_field_one: !str \"brave\"}}}",
                         "{nest_one: !obj {field_two_nest: !obj {nested_field_two: !str \"new\"}, field_three: !str \"world\"}}",
                         "{nest_two: !obj {field_one: !str \"world\", field_two: !str \"wide\"}}"],
                         SplitSisl(100).split_sisl(test_json))

    def test_list_of_different_types_is_converted_to_sisl_and_split(self):
        self.assertEqual(["{nest_one: !list {_0: !int \"1\", _1: !str \"string1\", _2: !bool \"true\", _3: !list {_0: !list {_0: !str \"nested_list\"}}}}",
                         "{nest_one: !list {_3: !list {_0: !list {_1: !int \"2\"}, _1: !bool \"false\", _2: !float \"3.0\"}}}",
                         "{nest_one: !list {_4: !obj {field_one: !bool \"false\", field_two: !str \"string3\"}}}",
                         "{nest_two: !obj {field_one: !str \"string4\", sub_nest_one: !obj {field_two: !obj {field_three: !float \"4.0\"}}}}",
                         "{nest_two: !obj {field_four: !bool \"true\", sub_nest_two: !obj {field_five: !int \"3\"}}}"],
                         SplitSisl(120).split_sisl({"nest_one": [1, "string1", True, [["nested_list", 2], False, 3.0],
                                                                 {"field_one": False, "field_two": "string3"}],
                                                   "nest_two": {"field_one": "string4", "sub_nest_one": {"field_two": {"field_three": 4.0}}, "field_four": True, "sub_nest_two": {"field_five": 3}}}))

    def test_list_of_nested_dicts_converted_sisl(self):
        list = "{list_one: !list {"
        item_0 = "_0: !str \"item_0\", "
        item_1 = "_1: !str \"item_1\", "
        item_2 = "_2: !obj {item_2: !obj {nest_field_1: !str \"nest_item_1_value\"}}, "
        item_3_part1 = "_3: !obj {item_3: !obj {nest_field_1: !obj {nest_field_2: !obj {nest_field_3: !obj {nest_field_4: !obj {nest_field_5: !str \"nest_field_5_value\"}}}}}}"
        item_3_part2 = "_3: !obj {item_3: !obj {nest_field_1: !obj {nest_field_2: !obj {nest_field_3: !obj {nest_field_4_key_2: !obj {nest_field_5_key_2: !str \"nest_field_5_value_2\"}}}}}}, "
        item_4_part1 = "_4: !obj {item_4: !obj {nest_field_1: !obj {nest_field_2: !str \"nest_field_2_value\"}}}"
        item_4_part2 = "_4: !obj {item_4: !obj {nest_field_1_key_2: !str \"nest_item_2\"}, nest_field_2_key_2: !obj {nest_field_3_key_2: !str \"nest_field_3_key_2_value_2\"}}"

        line_1 = list + item_0 + item_1 + item_2 + item_3_part1 + "}}"
        line_2 = list + item_3_part2 + item_4_part1 + "}}"
        line_3 = list + item_4_part2 + "}}"

        item_2_dict = {"item_2": {"nest_field_1": "nest_item_1_value"}}
        item_3_dict = {"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}
        item_4_dict = {"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                  "nest_field_1_key_2": "nest_item_2"},
                       "nest_field_2_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}

        self.assertEqual([line_1, line_2, line_3],
                         SplitSisl(300).split_sisl({"list_one": ["item_0", "item_1", item_2_dict, item_3_dict, item_4_dict]}))


if __name__ == '__main__':
    unittest.main()





# Copyright PA Knowledge Ltd 2021

import unittest

from pysisl.sisl_joiner import JoinSisl


class SislJoinerTests(unittest.TestCase):
    def test_join_list_with_empty_value(self):
        list_of_dicts = [{"field_one": {"type": "type", "value": {}}}]
        expected_output = {"field_one": {"type": "type", "value": {}}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_list_with_only_one_dict(self):
        list_of_dicts = [{"field_one": {"type": "type", "value": "value1"}}]
        expected_output = {"field_one": {"type": "type", "value": "value1"}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_list_of_multiple_dicts(self):
        first_dict = {"abc": {"type": "type", "value": "value1"}}
        second_dict = {"def": {"type": "type", "value": "value2"}}
        third_dict = {"ghi": {"type": "type", "value": "value3"}}
        expected_output = {'abc': {'type': 'type', 'value': 'value1'},
                           'def': {'type': 'type', 'value': 'value2'},
                           "ghi": {"type": "type", "value": "value3"}}

        self.assertEqual(expected_output, JoinSisl.join([first_dict, second_dict, third_dict]))

    def test_join_list_of_nested_dicts(self):
        list_of_dicts = [{"field_one": {"type": "type", "value": {"field_one_nested": {"type": "type", "value": "value1"}}}},
                         {"field_two": {"type": "type", "value": {"field_two_nested": {"type": "type", "value": "value2"}}}}]

        expected_output = {"field_one": {"type": "type", "value": {"field_one_nested": {"type": "type", "value": "value1"}}},
                           "field_two": {"type": "type", "value": {"field_two_nested": {"type": "type", "value": "value2"}}}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_simple_list_structure(self):
        list_of_dicts = [{"abc": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"}}}},
                         {"abc": {"type": "type", "value": {"_1": {"type": "type", "value": "value2"}}}}]

        expected_output = {"abc": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"},
                                                             "_1": {"type": "type", "value": "value2"}}}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_list_containing_separate_nested_list_elements(self):
        # [[1]], [[2]] -> [[1], [2]]
        list_of_dicts = [{"abc": {"type": "type", "value": {"_0": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"}}}}}},
                         {"abc": {"type": "type", "value": {"_1": {"type": "type", "value": {"_0": {"type": "type", "value": "value2"}}}}}}]

        expected_output = {"abc": {"type": "type", "value": {
            "_0": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"}}},
            "_1": {"type": "type", "value": {"_0": {"type": "type", "value": "value2"}}}
        }}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_nested_list_structure_that_is_split_in_nested_list(self):
        # [1, [2]], [[3], 4] -> [1, [2, 3], 4]
        list_of_dicts = [{"abc": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"},
                                                            "_1": {"type": "type", "value": {"_0": {"type": "type", "value": "value2"}}}}}},
                         {"abc": {"type": "type", "value": {"_1": {"type": "type", "value": {"_1": {"type": "type", "value": "value3"}}},
                                                            "_2": {"type": "type", "value": "value4"}}}}]

        expected_output = {"abc": {"type": "type", "value": {"_0": {"type": "type", "value": "value1"},
                                                             "_1": {"type": "type", "value": {"_0": {"type": "type", "value": "value2"},
                                                                                              "_1": {"type": "type", "value": "value3"}}},
                                                             "_2": {"type": "type", "value": "value4"}}}}

        self.assertEqual(expected_output, JoinSisl.join(list_of_dicts))

    def test_join_deeply_nested_structure_with_key_value_pairs_at_multiple_different_levels(self):
        # {"field_one": {"field_two_one": {"field_three": {"field_four": "value1"}}}},
        # {"field_one": {"field_two_one": {"field_three": {"field_five": "value2"}}}},
        # {"field_one": {"field_two_two": "value3"}} ->
        # {"field_one": {"field_two_one": {"field_three": {"field_four": "value1", "field_five": "value2"}},
        #                "field_two_two": "value3"}}

        first_dict = {"field_one": {"type": "type", "value": {"field_two_one": {"type": "type", "value": {"field_three": {"type": "type", "value": {"field_four": {"type": "type", "value": "value1"}}}}}}}}
        second_dict = {"field_one": {"type": "type", "value": {"field_two_one": {"type": "type", "value": {"field_three": {"type": "type", "value": {"field_five": {"type": "type", "value": "value2"}}}}}}}}
        third_dict = {"field_one": {"type": "type", "value": {"field_two_two": {"type": "type", "value": "value3"}}}}

        expected_output = {"field_one": {"type": "type", "value": {"field_two_one": {"type": "type", "value": {"field_three": {"type": "type", "value": {"field_four": {"type": "type", "value": "value1"},
                                                                                                                                                         "field_five": {"type": "type", "value": "value2"}}}}},
                                                                   "field_two_two": {"type": "type", "value": "value3"}}}}

        self.assertEqual(expected_output, JoinSisl.join([first_dict, second_dict, third_dict]))

    def test_list_of_nested_dicts_can_be_merged(self):
        # field_one: [dict_1, dict_2, dict_3]
        first_dict = {"field_one": {"type": "type", "value": {"field_one_nested": {"type": "type", "value": "value1"}}}}
        second_dict = {"field_one": {"type": "type", "value": "value2"}}
        third_dict = {"field_one": {"type": "type", "value": {"field_two": {"type": "type", "value": {"_0": {"type": "type", "value": "nested_item_0"},
                                                                                                      "_1": {"type": "type", "value": "nested_item_1"},
                                                                                                      "_2": {"type": "type", "value": "nested_item_2"}}}}}}

        input_list = [{"list_1": {"type": "type", "value": {"_0": {"type": "type", "value": first_dict}}}},
                      {"list_1": {"type": "type", "value": {"_1": {"type": "type", "value": second_dict}}}},
                      {"list_1": {"type": "type", "value": {"_2": {"type": "type", "value": third_dict}}}}]

        expected_output = {"list_1": {"type": "type", "value": {"_0": {"type": "type", "value": first_dict},
                                                                "_1": {"type": "type", "value": second_dict},
                                                                "_2": {"type": "type", "value": third_dict}}}}

        self.assertEqual(expected_output, JoinSisl.join(input_list))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_type_flatten import SISLTypeFlatten, UnsupportedSislType


class SislFlatteningTests(unittest.TestCase):
    def test_basic_sisl_structures(self):
        self.assertEqual({"abc": "stuff"}, SISLTypeFlatten.flatten({"abc": {"type": "str", "value": "stuff"}}))

    def test_anonymous_string(self):
        self.assertEqual("stuff", SISLTypeFlatten.flatten({"_": {"type": "_str", "value": "stuff"}}))

    def test_bool(self):
        self.assertEqual({"abc": True}, SISLTypeFlatten.flatten({"abc": {"type": "bool", "value": "true"}}))
        self.assertEqual({"abc": False}, SISLTypeFlatten.flatten({"abc": {"type": "bool", "value": "false"}}))
        self.assertRaises(UnsupportedSislType, SISLTypeFlatten.flatten, {"abc": {"type": "bool", "value": "eek"}})

    def test_integer(self):
        self.assertEqual({"abc": 1}, SISLTypeFlatten.flatten({"abc": {"type": "int", "value": "1"}}))
        self.assertEqual({"abc": -1}, SISLTypeFlatten.flatten({"abc": {"type": "int", "value": "-1"}}))
        self.assertEqual({"abc": 1000000000000000000000},
                         SISLTypeFlatten.flatten({"abc": {"type": "int", "value": "1000000000000000000000"}}))

    def test_integer_errors_when_decimal_provided(self):
        self.assertRaises(ValueError, SISLTypeFlatten.flatten, {"abc": {"type": "int", "value": "1.1"}})

    def test_integer_errors_if_string(self):
        self.assertRaises(ValueError, SISLTypeFlatten.flatten, {"abc": {"type": "int", "value": "eek"}})

    def test_float(self):
        self.assertEqual({"abc": 1.1}, SISLTypeFlatten.flatten({"abc": {"type": "float", "value": "1.1"}}))
        self.assertEqual({"abc": -1.1}, SISLTypeFlatten.flatten({"abc": {"type": "float", "value": "-1.1"}}))
        self.assertEqual({"abc": 1}, SISLTypeFlatten.flatten({"abc": {"type": "float", "value": "1.0"}}))
        self.assertEqual({"abc": 1}, SISLTypeFlatten.flatten({"abc": {"type": "float", "value": "1"}}))

    def test_empty_list(self):
        self.assertEqual({"abc": []}, SISLTypeFlatten.flatten({"abc": {"type": "list", "value": {}}}))

    def test_single_element_list(self):
        self.assertEqual({"abc": [1]}, SISLTypeFlatten.flatten({"abc": {"type": "list", "value": {"_0": {"type": "int", "value": "1"}}}}))

    def test_two_element_list(self):
        self.assertEqual({"abc": [1, 5]}, SISLTypeFlatten.flatten(
            {"abc": {"type": "list", "value": {
                "_0": {"type": "int", "value": "1"},
                "_1": {"type": "int", "value": "5"}
            }}}))

        self.assertEqual({"abc": [[1], [5]]}, SISLTypeFlatten.flatten(
            {"abc": {"type": "list", "value": {
                "_0": {"type": "list", "value": {"_0": {"type": "int", "value": "1"}}},
                "_1": {"type": "list", "value": {"_0": {"type": "int", "value": "5"}}}
            }}}))

    def test_two_element_list_is_sorted(self):
        self.assertEqual({"abc": [5, 1]}, SISLTypeFlatten.flatten(
            {"abc": {"type": "list", "value": {
                "_1": {"type": "int", "value": "1"},
                "_0": {"type": "int", "value": "5"}
            }}}))

    def test_nested_list(self):
        self.assertEqual({"abc": [1, [2, "a"]]}, SISLTypeFlatten.flatten(
            {"abc": {"type": "list", "value": {
                "_0": {"type": "int", "value": "1"},
                "_1": {"type": "list", "value": {"_0": {"type": "int", "value": "2"},
                                                 "_1": {"type": "str", "value": "a"}}}}}}))
        self.assertEqual({"abc": [1, [2, {"abc": 4}, {"fgh": [2, 3, 4, 5]}]]}, SISLTypeFlatten.flatten(
            {"abc": {"type": "list", "value": {
                "_0": {"type": "int", "value": "1"},
                "_1": {"type": "list", "value": {"_0": {"type": "int", "value": "2"},
                                                 "_1": {"type": "obj", "value": {"abc": {"type": "int", "value": "4"}}},
                                                 "_2": {"type": "obj", "value": {"fgh": {"type": "list", "value": {
                                                     "_0": {"type": "int", "value": "2"},
                                                     "_1": {"type": "int", "value": "3"},
                                                     "_2": {"type": "int", "value": "4"},
                                                     "_3": {"type": "int", "value": "5"}}}}}}}}}}))

    def test_single_element_anonymous_list(self):
        self.assertEqual([1], SISLTypeFlatten.flatten({"_": {"type": "_list", "value": {"_0": {"type": "int", "value": "1"}}}}))

    def test_two_element_anonymous_list(self):
        self.assertEqual([1, 5], SISLTypeFlatten.flatten(
            {"_": {"type": "_list", "value": {
                "_0": {"type": "int", "value": "1"},
                "_1": {"type": "int", "value": "5"}
            }}}))

        self.assertEqual([[1], [5]], SISLTypeFlatten.flatten(
            {"_": {"type": "_list", "value": {
                "_0": {"type": "list", "value": {"_0": {"type": "int", "value": "1"}}},
                "_1": {"type": "list", "value": {"_0": {"type": "int", "value": "5"}}}
            }}}))

    def test_none_type(self):
        self.assertEqual({"abc": None}, SISLTypeFlatten.flatten({"abc": {"type": "null", "value": ""}}))

    def test_float_errors_if_string(self):
        self.assertRaises(ValueError, SISLTypeFlatten.flatten, {"abc": {"type": "float", "value": "eek"}})

    def test_unsupported_type_does_nothing(self):
        self.assertEqual({"_abc": {"type": "not_supported", "value": "value"}},
                         SISLTypeFlatten.flatten({"_abc": {"type": "not_supported", "value": "value"}}))

    def test_nested_sisl_structures(self):
        self.assertEqual({"abc": {"def": "value"}},
                         SISLTypeFlatten.flatten(
                             {"abc": {"type": "obj", "value": {"def": {"type": "str", "value": "value"}}}}))

    def test_backslash_and_quotes_escaped(self):
        self.assertEqual({"key1": r'some string \ " message'},
                         SISLTypeFlatten.flatten({"key1": {"type": "str", "value": r'some string \\ \" message'}}))


if __name__ == '__main__':
    unittest.main()




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'BEGIN_OBJECT END_OBJECT LIST_SEPARATOR NAME TYPE VALUEsisl_object : BEGIN_OBJECT sisl_triplets END_OBJECT sisl_triplets : empty_triplet\n                     | sisl_triplet\n                     | sisl_triplet_list  empty_triplet :  sisl_triplet : NAME type_and_value  type_and_value : type value  type : TYPE  value : VALUE\n              | sisl_object  sisl_triplet_list : sisl_triplet LIST_SEPARATOR sisl_triplets '
    
_lr_action_items = {'BEGIN_OBJECT':([0,11,12,],[2,2,-8,]),'$end':([1,8,],[0,-1,]),'END_OBJECT':([2,3,4,5,6,8,9,10,13,14,15,16,],[-5,8,-2,-3,-4,-1,-5,-6,-11,-7,-9,-10,]),'NAME':([2,9,],[7,7,]),'LIST_SEPARATOR':([5,8,10,14,15,16,],[9,-1,-6,-7,-9,-10,]),'TYPE':([7,],[12,]),'VALUE':([11,12,],[15,-8,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'sisl_object':([0,11,],[1,16,]),'sisl_triplets':([2,9,],[3,13,]),'empty_triplet':([2,9,],[4,4,]),'sisl_triplet':([2,9,],[5,5,]),'sisl_triplet_list':([2,9,],[6,6,]),'type_and_value':([7,],[10,]),'type':([7,],[11,]),'value':([11,],[14,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> sisl_object","S'",1,None,None,None),
  ('sisl_object -> BEGIN_OBJECT sisl_triplets END_OBJECT','sisl_object',3,'p_sisl_object','sisl_parser.py',7),
  ('sisl_triplets -> empty_triplet','sisl_triplets',1,'p_sisl_triplets','sisl_parser.py',12),
  ('sisl_triplets -> sisl_triplet','sisl_triplets',1,'p_sisl_triplets','sisl_parser.py',13),
  ('sisl_triplets -> sisl_triplet_list','sisl_triplets',1,'p_sisl_triplets','sisl_parser.py',14),
  ('empty_triplet -> <empty>','empty_triplet',0,'p_empty_triplet','sisl_parser.py',19),
  ('sisl_triplet -> NAME type_and_value','sisl_triplet',2,'p_sisl_triplet','sisl_parser.py',24),
  ('type_and_value -> type value','type_and_value',2,'p_type_and_value','sisl_parser.py',29),
  ('type -> TYPE','type',1,'p_type','sisl_parser.py',34),
  ('value -> VALUE','value',1,'p_value','sisl_parser.py',39),
  ('value -> sisl_object','value',1,'p_value','sisl_parser.py',40),
  ('sisl_triplet_list -> sisl_triplet LIST_SEPARATOR sisl_triplets','sisl_triplet_list',3,'p_sisl_triplet_list','sisl_parser.py',45),
]



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_encoder import SislEncoder
from pysisl.sisl_decoder import SislDecoder


class SislEncoderDecoderTests(unittest.TestCase):
    def test_empty_dict(self):
        encoder_input = {}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_string(self):
        encoder_input = {"field_one": "string"}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_underscore_key_string(self):
        encoder_input = {"_": "string"}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_anonymous_string(self):
        encoder_input = "string"
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_unicode_string(self):
        encoder_input = {"abc": "ӧӧФ@፨=😂❤🔥❌£¥"}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder().dumps(encoder_input)))

    def test_multiple_nested_dict(self):
        encoder_input = {"field_one": {"field_two": 123, "field_three": {"field_four": "field_four string"}}}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_float(self):
        encoder_input = {"field_one": 1.0}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_anonymous_bool(self):
        self.assertEqual(True, SislDecoder().loads(SislEncoder.dumps(True)))

    def test_anonymous_int(self):
        self.assertEqual(1, SislDecoder().loads(SislEncoder.dumps(1)))

    def test_nested_list_in_value(self):
        encoder_input = {"field_one": [1, [2, 3]]}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_anonymous_nested_list(self):
        encoder_input = [1, [2, 3]]
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_single_trailing_comma(self):
        encoder_input = {"abc": 1,}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_escaped_backslash_in_value(self):
        encoder_input = {"abc": "te\\st"}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_nested_dict_structure_with_lists_in_value(self):
        encoder_input = {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"},
                                                  "lvl2_2": [["value_8"],
                                                             {"lvl3": {"lvl4": "value_9"}},
                                                             {"lvl3": "value_10"}]},
                                         "lvl1_3": {"lvl2_1": {"lvl3_1": {"lvl4_1": "value_12"}}}}}
        self.assertEqual(encoder_input, SislDecoder().loads(SislEncoder.dumps(encoder_input)))

    def test_decoded_sisl_with_escaped_backslash(self):
        decoder_input = r'{abc: !str "te\\st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quotes(self):
        decoder_input = r'{abc: !str "te\"st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_backslash_before_closing_quote(self):
        decoder_input = r'{abc: !str "test\\"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quote_before_closing_quote(self):
        decoder_input = r'{abc: !str "test\""}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quote_before_comma(self):
        decoder_input = r'{abc: !str "te\"a,st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quote_just_before_comma(self):
        decoder_input = r'{abc: !str "te\",st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quote_before_space_comma(self):
        decoder_input = r'{abc: !str "te\" ,st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_escaped_quote_before_brace(self):
        decoder_input = r'{abc: !str "te\"}st"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))

    def test_decoded_sisl_with_unicode_string(self):
        decoder_input = r'{abc: !str "\u04e7\u04e7\u0424@\u1368=\U0001f602b\u2764\U0001f525\u274c\xa3\xa5"}'
        self.assertEqual(decoder_input, SislEncoder.dumps(SislDecoder().loads(decoder_input)))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_decoder import SislDecoder, SislValidationError


class SislParsingTests(unittest.TestCase):
    def test_basic_sisl_structures(self):
        self.assertEqual({"abc": {"results": {"result": 123}}}, SislDecoder().loads("{abc: !obj {results: !obj {result: !int \"123\"}}}"))

    def test_successful_parse_with_schema(self):
        schema = {
            "properties": {
                "field_one": {
                    "type": "string"
                },
                "field_two": {
                    "type": "number"
                }
            }
        }
        self.assertEqual({"field_one": "string", "field_two": 2},
                         SislDecoder().loads("{field_one: !str \"string\", field_two: !int \"2\"}", schema=schema))

    def test_parse_with_schema_fails_throws_error(self):
        schema = {
            "properties": {
                "field_one": {
                    "type": "number"
                },
                "field_two": {
                    "type": "number"
                }
            }
        }
        self.assertRaises(SislValidationError,
                          SislDecoder().loads, "{field_one: !str \"string\", field_two: !int \"2\"}", schema=schema)

    def test_sisl_with_escaped_backslash(self):
        self.assertEqual({"abc": "te\\st"}, SislDecoder().loads("{abc: !str \"te\\st\"}"))

    def test_sisl_with_escaped_quotes(self):
        self.assertEqual({"abc": "te\"st"}, SislDecoder().loads(r'{abc: !str "te\"st"}'))

    def test_escaped_backslash_before_closing_quote_sisl(self):
        self.assertEqual({"abc": "test\\"}, SislDecoder().loads(r'{abc: !str "test\\"}'))

    def test_escaped_quote_before_closing_quote_sisl(self):
        self.assertEqual({"abc": "test\""}, SislDecoder().loads(r'{abc: !str "test\""}'))

    def test_sisl_with_8_bit_unicode(self):
        self.assertEqual({"abc": "çä"}, SislDecoder().loads(r'{abc: !str "\xe7\xe4"}'))

    def test_sisl_with_16_bit_unicode(self):
        self.assertEqual({"abc": "ӧʶ"}, SislDecoder().loads(r'{abc: !str "\u04e7\u02b6"}'))

    def test_sisl_with_32_bit_unicode(self):
        self.assertEqual({"abc": "😂🔥"}, SislDecoder().loads(r'{abc: !str "\U0001f602\U0001f525"}'))

    def test_sisl_with_anonymous_string(self):
        self.assertEqual("i_am_a_string", SislDecoder().loads(r'{_: !_str "i_am_a_string"}'))

    def test_sisl_with_anonymous_unicode_string(self):
        self.assertEqual("😂", SislDecoder().loads(r'{_: !_str "\U0001f602"}'))

    def test_sisl_with_anonymous_list(self):
        self.assertEqual([1,2,3], SislDecoder().loads(['{_: !_list {_0: !int "1", _1: !int "2", _2: !int "3"}}']))

    def test_basic_split_sisl_string_parsed_using_joiner(self):
        self.assertEqual({"abc": 2, "def": 3}, SislDecoder().loads(['{abc: !int "2"}', '{def: !int "3"}']))

    def test_split_nested_dict_rejoined(self):
        self.assertEqual({"field_one": {"field_two": 123, "field_three": {"field_four": "field_four string"}}},
                         SislDecoder().loads(['{field_one: !obj {field_two: !int "123"}}', '{field_one: !obj {field_three: !obj {field_four: !str "field_four string"}}}']))

    def test_split_deeply_nested_dict_rejoined(self):
        input_sisl = ["{field_one: !obj {field_two: !obj {field_three: !obj {field_four: !str \"value_1\"}}}}",
                      "{field_one: !obj {field_two: !obj {field_three: !obj {field_five: !obj {field_six: !obj {field_seven: !obj {field_eight: !str \"value_2\"}}}}}}}",
                      "{field_one: !obj {field_two: !obj {field_three: !obj {field_five: !obj {field_six: !obj {field_seven: !obj {field_nine: !str \"value_3\"}}}}}}}",
                      "{field_one: !obj {field_two: !obj {field_ten: !str \"value_4\"}}}"]
        expected_output = {"field_one": {"field_two": {"field_three": {"field_four": "value_1",
                                                                       "field_five": {"field_six": {"field_seven": {"field_eight": "value_2", "field_nine": "value_3"}}}},
                                                       "field_ten": "value_4"}}}
        self.assertEqual(expected_output,
                         SislDecoder().loads(input_sisl))

    def test_split_nested_dict_with_long_values_rejoined(self):
        expected_output = {"nest_one": {"sub_nest_one": {"nested_field_one": "test 1", "nested_field_two": "testing this long string still fits"}}}
        input_sisl = ["{nest_one: !obj {sub_nest_one: !obj {nested_field_one: !str \"test 1\"}}}",
                      "{nest_one: !obj {sub_nest_one: !obj {nested_field_two: !str \"testing this long string still fits\"}}}"]
        self.assertEqual(expected_output, SislDecoder().loads(input_sisl))

    def test_basic_list_that_is_parsed_using_joiner(self):
        self.assertEqual({"field_one": ["I", "am", "a", "list"]},
                         SislDecoder().loads(["{field_one: !list {_0: !str \"I\", _1: !str \"am\"}}", "{field_one: !list {_2: !str \"a\", _3: !str \"list\"}}"]))

    def test_basic_split_nested_list_is_parsed_using_joiner(self):
        self.assertEqual({"field_one": ["I", ["am", "a"], "list"]},
                         SislDecoder().loads(["{field_one: !list {_0: !str \"I\", _1: !list {_0: !str \"am\"}}}", "{field_one: !list {_1: !list {_1: !str \"a\"}, _2: !str \"list\"}}"]))

    def test_split_list_of_different_types_is_rejoined(self):
        input_sisl = ["{nest_one: !list {_0: !int \"1\", _1: !str \"string1\", _2: !bool \"true\", _3: !list {_0: !list {_0: !str \"nested_list\"}}}}",
                      "{nest_one: !list {_3: !list {_0: !list {_1: !int \"2\"}, _1: !bool \"false\", _2: !float \"3.0\"}}}",
                      "{nest_one: !list {_4: !obj {field_one: !bool \"false\", field_two: !str \"string3\"}}}",
                      "{nest_two: !obj {field_one: !str \"string4\", sub_nest_one: !obj {field_two: !obj {field_three: !float \"4.0\"}}}}",
                      "{nest_two: !obj {field_four: !bool \"true\", sub_nest_two: !obj {field_five: !int \"3\"}}}"]
        expected_output = {"nest_one": [1, "string1", True, [["nested_list", 2], False, 3.0],
                                        {"field_one": False, "field_two": "string3"}],
                           "nest_two": {"field_one": "string4", "sub_nest_one": {"field_two": {"field_three": 4.0}}, "field_four": True, "sub_nest_two": {"field_five": 3}}}
        self.assertEqual(expected_output, SislDecoder().loads(input_sisl))

    def test_complex_list_of_nested_dicts_that_is_split_is_parsed_using_joiner(self):
        list = "{list_one: !list {"
        item_0 = "_0: !str \"item_0\", "
        item_1 = "_1: !str \"item_1\", "
        item_2 = "_2: !obj {item_2: !obj {nest_field_1: !str \"nest_item_1_value\"}}, "
        item_3_part1 = "_3: !obj {item_3: !obj {nest_field_1: !obj {nest_field_2: !obj {nest_field_3: !obj {nest_field_4: !obj {nest_field_5: !str \"nest_field_5_value\"}}}}}}"
        item_3_part2 = "_3: !obj {item_3: !obj {nest_field_1: !obj {nest_field_2: !obj {nest_field_3: !obj {nest_field_4_key_2: !obj {nest_field_5_key_2: !str \"nest_field_5_value_2\"}}}}}}, "
        item_4_part1 = "_4: !obj {item_4: !obj {nest_field_1: !obj {nest_field_2: !str \"nest_field_2_value\"}}}"
        item_4_part2 = "_4: !obj {item_4: !obj {nest_field_1_key_2: !str \"nest_item_2\"}, nest_field_2_key_2: !obj {nest_field_3_key_2: !str \"nest_field_3_key_2_value_2\"}}"

        line_1 = list + item_0 + item_1 + item_2 + item_3_part1 + "}}"
        line_2 = list + item_3_part2 + item_4_part1 + "}}"
        line_3 = list + item_4_part2 + "}}"

        item_2_dict = {"item_2": {"nest_field_1": "nest_item_1_value"}}
        item_3_dict = {"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}
        item_4_dict = {"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                  "nest_field_1_key_2": "nest_item_2"},
                       "nest_field_2_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}

        input_sisl = [line_1, line_2, line_3]
        expected_output = {"list_one": ["item_0", "item_1", item_2_dict, item_3_dict, item_4_dict]}
        self.assertEqual(expected_output, SislDecoder().loads(input_sisl))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_encoder import SislEncoder, TypeValidationError
from pysisl.deep_merger import ListTuple


class NotSislClass:
    def __init__(self):
        self.name = "unknown object"


class SislParsingTests(unittest.TestCase):
    def test_basic_dict_structure(self):
        self.assertEqual("{}", SislEncoder.dumps({}))

    def test_string_values(self):
        self.assertEqual("{field_one: !str \"string\"}", SislEncoder.dumps({"field_one": "string"}))
        self.assertEqual("{field_one: !str \"string\", field_two: !str \"string2\"}",
                         SislEncoder.dumps({"field_one": "string", "field_two": "string2"}))

    def test_anonymous_string_values(self):
        self.assertEqual("{_: !_str \"string\"}", SislEncoder.dumps("string"))

    def test_int_values(self):
        self.assertEqual("{field_one: !int \"123\"}", SislEncoder.dumps({"field_one": 123}))

    def test_float_values(self):
        self.assertEqual("{field_one: !float \"1.23\"}", SislEncoder.dumps({"field_one": 1.23}))
        self.assertEqual("{field_one: !float \"-0.123\"}", SislEncoder.dumps({"field_one": -0.123}))
        self.assertEqual("{field_one: !float \"1.0\"}", SislEncoder.dumps({"field_one": 1.0}))

    def test_bool_values(self):
        self.assertEqual("{field_one: !bool \"true\"}", SislEncoder.dumps({"field_one": True}))
        self.assertEqual("{field_one: !bool \"false\"}", SislEncoder.dumps({"field_one": False}))

    def test_single_element_integer_list_values(self):
        self.assertEqual("{field_one: !list {_0: !int \"1\"}}", SislEncoder.dumps({"field_one": [1]}))

    def test_list_values(self):
        self.assertEqual("{field_one: !list {_0: !int \"1\", _1: !int \"2\", _2: !int \"3\"}}", SislEncoder.dumps({"field_one": [1, 2, 3]}))
        self.assertEqual("{field_one: !list {_0: !list {_0: !int \"4\"}, _1: !int \"2\", _2: !int \"3\"}}", SislEncoder.dumps({"field_one": [[4], 2, 3]}))

    def test_anonymous_list_values(self):
        self.assertEqual("{_: !_list {_0: !int \"1\", _1: !int \"2\", _2: !int \"3\"}}", SislEncoder.dumps([1, 2, 3]))
        self.assertEqual("{_: !_list {_0: !list {_0: !int \"4\"}, _1: !int \"2\", _2: !int \"3\"}}", SislEncoder.dumps([[4], 2, 3]))
        self.assertEqual("{_: !_list {_0: !obj {test: !str \"string\"}, _1: !int \"2\", _2: !int \"3\"}}", SislEncoder.dumps([{"test":"string"}, 2, 3]))

    def test_basic_nested_dict_structure(self):
        self.assertEqual("{field_one: !obj {}}", SislEncoder.dumps({"field_one": {}}))

    def test_nested_dict_structure(self):
        self.assertEqual("{field_one: !obj {field_two: !int \"123\"}}", SislEncoder.dumps({"field_one": {"field_two": 123}}))

    def test_multiple_nested_dict_structure(self):
        self.assertEqual("{field_one: !obj {field_two: !int \"123\", field_three: !str \"field_three string\"}}",
                         SislEncoder.dumps({"field_one": {"field_two": 123, "field_three": "field_three string"}}))
        self.assertEqual("{field_one: !obj {field_two: !int \"123\", field_three: !obj {field_four: !str \"field_four string\"}}}",
                         SislEncoder.dumps({"field_one": {"field_two": 123, "field_three": {"field_four": "field_four string"}}}))

    def test_unknown_types_in_dictionary_should_throw_error(self):
        self.assertRaises(TypeValidationError, SislEncoder.dumps, {"field_two": NotSislClass()})

    def test_none_type(self):
        self.assertEqual("{field_one: !null \"\"}", SislEncoder.dumps({"field_one": None}))

    def test_escaped_backslash_successful(self):
        self.assertEqual(r'{field_one: !str "te\\st"}', SislEncoder.dumps({"field_one": "te\\st"}))

    def test_escaped_dbquote_successful(self):
        self.assertEqual(r'{field_one: !str "te\"st"}', SislEncoder.dumps({"field_one": "te\"st"}))

    def test_8_bit_unicode_characters(self):
        self.assertEqual(r'{field_one: !str "\xe7\xe4"}', SislEncoder.dumps({"field_one": "çä"}))

    def test_16_bit_unicode_characters(self):
        self.assertEqual(r'{field_one: !str "\u04e7\u02b6"}', SislEncoder.dumps({"field_one": "ӧʶ"}))

    def test_32_bit_unicode_characters(self):
        self.assertEqual(r'{field_one: !str "\U0001f602\U0001f525"}', SislEncoder.dumps({"field_one": "😂🔥"}))

    def test_list_tuple_value(self):
        # [value1, value2]
        self.assertEqual("{field_one: !list {_0: !str \"value1\", _1: !str \"value2\"}}",
                         SislEncoder.dumps({"field_one": [ListTuple("value1", 0), ListTuple("value2", 1)]}))

    def test_nested_list_tuple(self):
        # [[value1]]
        self.assertEqual("{field_one: !list {_0: !list {_0: !str \"value1\"}}}",
                         SislEncoder.dumps({"field_one": [ListTuple([ListTuple("value1", 0)], 0)]}))

    def test_deeply_nested_list_tuple(self):
        # {"field_one": [value1, [[[value2]]]]}
        self.assertEqual("{field_one: !list {_0: !str \"value1\", _1: !list {_0: !list {_0: !list {_0: !str \"value2\"}}}}}",
                         SislEncoder.dumps({"field_one": [ListTuple("value1", 0), ListTuple([ListTuple([ListTuple([ListTuple("value2", 0)], 0)], 0)], 1)]}))

    def test_list_of_dicts(self):
        # {"list_one": [{"dict_1": "value_1"}, {"dict_2": "value_2"}, {'dict_3': 'value_3'}]
        self.assertEqual("{list_one: !list {_0: !obj {dict_1: !str \"value_1\"}, _1: !obj {dict_2: !str \"value_2\"}, _2: !obj {dict_3: !str \"value_3\"}}}",
                         SislEncoder.dumps({"list_one": [ListTuple({"dict_1": "value_1"}, 0),
                                                         ListTuple({"dict_2": "value_2"}, 1),
                                                         ListTuple({'dict_3': 'value_3'}, 2)]}))

    def test_list_of_nested_items(self):
        # {"field_one": ["A",
        #               [[["nested", "list"]]],
        #               {"lvl1": {"lvl2": {"item_1": "nested",
        #                                  "item_2": "dict"}}}
        #               ]}
        input_dict = {"field_one": [ListTuple("A", 0),
                                    ListTuple([ListTuple([ListTuple([ListTuple("nested", 0), ListTuple("list", 1)], 0)], 0)], 1),
                                    ListTuple({"lvl1": {"lvl2": {"item_1": "nested", "item_2": "dict"}}}, 2)]}

        self.assertEqual("{field_one: !list {_0: !str \"A\", "
                                            "_1: !list {_0: !list {_0: !list {_0: !str \"nested\", _1: !str \"list\"}}}, "
                                            "_2: !obj {lvl1: !obj {lvl2: !obj {item_1: !str \"nested\", "
                                                                              "item_2: !str \"dict\"}}}}}",
                         SislEncoder.dumps(input_dict))

    def test_list_of_nested_items_with_dicts_and_lists(self):
        # {"nested_list": [{"item1": {"nest_field_1": "nest_item_1_value"}}, ["item5", {"item5_2": "value_2"}]]}
        input_dict = {"nested_list": [ListTuple({"item1": {"nest_field_1": "nest_item_1_value"}}, 0),
                                      ListTuple([ListTuple("item5", 0),
                                                 ListTuple({"item5_2": "value_2"}, 1)], 1)]}

        self.assertEqual("{nested_list: !list {_0: !obj {item1: !obj {nest_field_1: !str \"nest_item_1_value\"}}, "
                                              "_1: !list {_0: !str \"item5\", "
                                                         "_1: !obj {item5_2: !str \"value_2\"}}}}",
                         SislEncoder.dumps(input_dict))

    def test_nested_dict_that_contains_nested_list(self):
        # {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"},
        #                                       "lvl2_2": [["value_8"],
        #                                                  {"lvl3": {"lvl4": "value_9"}},
        #                                                  {"lvl3": "value_10"}]},
        #                              "lvl1_3": {"lvl2_1": {"lvl3_1": {"lvl4_1": "value_12"}}}}}
        input_dict = {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"},
                                               "lvl2_2": [ListTuple([ListTuple("value_8", 0)], 0),
                                                          ListTuple({"lvl3": {"lvl4": "value_9"}}, 1),
                                                          ListTuple({"lvl3": "value_10"}, 2)]},
                                      "lvl1_3": {"lvl2_1": {"lvl3_1": {"lvl4_1": "value_12"}}}}}

        self.assertEqual("{nested_dict: !obj {lvl1: !obj {lvl2_1: !obj {lvl3: !str \"value_6\"}, "
                                                         "lvl2_2: !list {_0: !list {_0: !str \"value_8\"}, "
                                                                        "_1: !obj {lvl3: !obj {lvl4: !str \"value_9\"}}, "
                                                                        "_2: !obj {lvl3: !str \"value_10\"}}"
                                                        "}, "
                                             "lvl1_3: !obj {lvl2_1: !obj {lvl3_1: !obj {lvl4_1: !str \"value_12\"}}}}}",
                         SislEncoder.dumps(input_dict))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021
import construct
from itertools import cycle
import secrets
from binascii import unhexlify


class SislWrapper:
    def __init__(self, key_generator=None):
        self.majorVersion = 1
        self.minorVersion = 0
        self.headerLen = 48
        self.encapsulationType = 1
        self.encapsulationConfig = 3
        self.encapsulationDataLen = 8
        self.key = SislWrapper.key_generation() if key_generator is None else key_generator
        self.padding = 0
        self.cloaked_dagger = self._build_cloaked_dagger_header()
        self.chunk_count = 0

    def wraps(self, data: str):
        return self.cloaked_dagger + SislWrapper._xor_data(self.key, SislWrapper._is_bytes(data))

    @staticmethod
    def _xor_data(key, data):
        return bytes([a ^ b for (a, b) in zip(data, cycle(key))])

    @staticmethod
    def _is_bytes(data):
        if isinstance(data, (bytes, bytearray)):
            return data
        else:
            raise TypeError("Input must be bytes")

    def unwraps(self, data):
        payload = self._get_payload(data)
        unwrapped = SislWrapper._xor_data(self.key, payload)
        self.chunk_count += 1
        return unwrapped

    def _get_payload(self, data):
        if self.chunk_count == 0:
            CloakDagger.check_valid_cloaked_dagger_header(SislWrapper._is_bytes(data))
            self.key = CloakDagger.get_cloak_dagger_field_by_name(data, 'Encap_Mask')
            return SislWrapper._extract_payload_from_data(data)
        else:
            return data

    @staticmethod
    def _extract_payload_from_data(data):
        return data[CloakDagger.get_cloak_dagger_field_by_name(data, 'Length'):]

    def _build_cloaked_dagger_header(self):
        return CloakDagger.cloak_dagger_bytes().build(
            dict(Major_Version=self.majorVersion,
                 Minor_Version=self.minorVersion,
                 Length=self.headerLen,
                 Encap_Type=self.encapsulationType,
                 Encap_Config=self.encapsulationConfig,
                 Encap_Dlen=self.encapsulationDataLen,
                 Encap_Mask=self.key,
                 Padding=self.padding))

    @staticmethod
    def key_generation():
        key = unhexlify(secrets.token_hex(8))
        if b'\x00' in key:
            key = SislWrapper.key_generation()
        return key


class CloakDagger:
    @staticmethod
    def cloak_dagger_bytes():
        return construct.Struct("Magic_Number_1" / construct.Const(b'\xd1\xdf\x5f\xff'),
                                "Major_Version" / construct.Int16ub,
                                "Minor_Version" / construct.Int16ub,
                                "Length" / construct.Int32ub,
                                "Encap_Type" / construct.Int32ub,
                                "Encap_Config" / construct.Int16ub,
                                "Encap_Dlen" / construct.Int16ub,
                                "Encap_Mask" / construct.Bytes(8),
                                "RESERVED" / construct.Padding(16),
                                "Magic_Number_2" / construct.Const(b'\xff\x5f\xdf\xd1')
                                )

    @staticmethod
    def check_valid_cloaked_dagger_header(data):
        try:
            CloakDagger.cloak_dagger_bytes().parse(data)
        except construct.core.ConstructError as e:
            raise InvalidHeaderError(f"Header could not be validated: {e}")

    @staticmethod
    def get_cloak_dagger_field_by_name(data, field):
        return CloakDagger.cloak_dagger_bytes().parse(data).search(field)


class InvalidHeaderError(Exception):
    pass



# Copyright PA Knowledge Ltd 2021

from deepmerge import Merger
from collections import namedtuple
import copy
ListTuple = namedtuple('ListTuple', ['value', 'position'])


class DeepMerger:
    @classmethod
    def merge(cls, base, nxt):
        base_copy = copy.deepcopy(base[-1])
        deep_merger = Merger([(list, cls.list_strategy), (dict, ["merge"])], ["override"], ["override"])
        deep_merger.merge(base_copy, nxt)
        return base_copy

    @classmethod
    def list_strategy(cls, config, path, base, nxt):
        list_merger = Merger([(list, cls.list_strategy), (dict, ["merge"])], ["override"], ["override"])

        if nxt[0].position == base[-1].position:
            cls.merge_list_tuples(base, nxt, list_merger)
        else:
            base.append(nxt[0])

        return base

    @classmethod
    def merge_list_tuples(cls, base, nxt, list_tuple_merger):
        if type(base[-1].value) is dict and type(nxt[0].value) is dict:
            list_tuple_merger.merge(base[-1].value, nxt[0].value)
        else:
            # remove outer ListTuple
            base = base[-1].value
            nxt = nxt[0].value

            if nxt[0].position == base[-1].position:
                cls.merge_list_tuples(base, nxt, list_tuple_merger)
            else:
                base.append(nxt[0])

        return base




# Copyright PA Knowledge Ltd 2021

class UnsupportedSislType(Exception):
    pass


class SISLTypeFlatten:
    @classmethod
    def flatten(cls, sisl_dict):
        for (key, type_value) in sisl_dict.items():
            if key == "_" and "_" in type_value["type"]:
                return cls._lookup_type_cast(type_value)
        return {key: cls._lookup_type_cast(type_value) for key, type_value in sisl_dict.items()}


    @classmethod
    def _lookup_type_cast(cls, type_value):
        return cls._type_cast_mappings().get(type_value["type"], lambda tvp: tvp)(type_value)

    @classmethod
    def _type_cast_mappings(cls):
        return {
            "str": cls._convert_string,
            "_str": cls._convert_string,
            "obj": lambda type_value: cls.flatten(type_value["value"]),
            "_obj": lambda type_value: cls.flatten(type_value["value"]),
            "bool": cls._convert_bool,
            "_bool": cls._convert_bool,
            "int": lambda type_value: int(type_value["value"]),
            "_int": lambda type_value: int(type_value["value"]),
            "float": lambda type_value: float(type_value["value"]),
            "_float": lambda type_value: float(type_value["value"]),
            "list": lambda type_value: cls._flatten_to_list(type_value["value"]),
            "_list": lambda type_value: cls._flatten_to_list(type_value["value"]),
            "null": lambda type_value: None,
            "_null": lambda type_value: None
        }

    @classmethod
    def _flatten_to_list(cls, sisl_dict):
        test = [cls._lookup_type_cast(value) for key, value in sorted(sisl_dict.items())]
        return test

    @staticmethod
    def _convert_string(type_value):
        return str(type_value['value']).replace(r'\"', r'"').encode().decode('unicode_escape')

    @staticmethod
    def _convert_bool(type_value):
        if type_value["value"] == "true":
            return True
        elif type_value["value"] == "false":
            return False
        else:
            raise UnsupportedSislType(type_value)



# Copyright PA Knowledge Ltd 2021

from . import sisl_decoder
from . import sisl_encoder
from . import sisl_splitter
from . import sisl_joiner
from . import sisl_wrapper


def loads(sisl, schema=None):
    return sisl_decoder.SislDecoder().loads(sisl, schema)


def dumps(dict_to_encode, **kwargs):
    if kwargs.get('max_length'):
        return sisl_splitter.SplitSisl(kwargs.get('max_length')).split_sisl(dict_to_encode)
    else:
        return sisl_encoder.SislEncoder.dumps(dict_to_encode)


def wraps(data):
    return sisl_wrapper.SislWrapper().wraps(data)


def unwraps(data):
    return sisl_wrapper.SislWrapper().unwraps(data)


class SislWrapper:
    def __init__(self):
        self.sisl = sisl_wrapper.SislWrapper()

    def wraps(self, data):
        return self.sisl.wraps(data)

    def unwraps(self, data):
        return self.sisl.unwraps(data)



# Copyright PA Knowledge Ltd 2021

from pysisl.sisl_encoder import SislEncoder
from pysisl.explode_structure import ExplodeStructure
from pysisl.remove_data import RemoveData


class SplitSisl:
    def __init__(self, max_length=1000000):
        self.max_length = max_length
        self.exploded_input = None

    def split_sisl(self, decoded_sisl_input):
        sisl_file = []
        self.exploded_input = self._convert_to_list_tuple(decoded_sisl_input)
        sisl_file.append(self._generate_sisl_file())
        while self._check_for_next_file_to_generate():
            sisl_file.append(self._generate_sisl_file())
        return sisl_file

    @staticmethod
    def _convert_to_list_tuple(decoded_sisl_input):
        return ExplodeStructure.explode_structure(decoded_sisl_input)[-1]

    def _generate_sisl_file(self):
        potential_sisl_file_candidates = ExplodeStructure.explode_structure(self.exploded_input)
        encoded_sisl, successful_struct = self._split_to_encoded_sisl(potential_sisl_file_candidates, self.max_length)
        self.exploded_input = RemoveData.subtract_data_from_input(self.exploded_input, successful_struct)
        return encoded_sisl

    def _check_for_next_file_to_generate(self):
        return not (self.exploded_input == {} or self.exploded_input == [])

    def _split_to_encoded_sisl(self, decoded_sisl_candidates, max_bytes):
        sisl = SislEncoder.dumps(decoded_sisl_candidates[0])
        self._check_sisl_length(max_bytes, sisl)
        decoded_sisl = decoded_sisl_candidates[0]

        for file in decoded_sisl_candidates:
            test_sisl = SislEncoder.dumps(file)
            if len(test_sisl.encode('utf-8')) > max_bytes:
                return sisl, decoded_sisl
            else:
                sisl = test_sisl
                decoded_sisl = file

        return sisl, decoded_sisl_candidates[-1]

    @staticmethod
    def _check_sisl_length(max_bytes, sisl):
        if len(sisl) > max_bytes:
            raise Exception(f"Unable to split input. {sisl} is greater than max size. ")



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.explode_structure import ExplodeStructure
from pysisl.deep_merger import ListTuple


class ExplodeStructureTests(unittest.TestCase):

    def test_explode_structure_for_flat_list(self):
        # {"field_one": ["I"]}
        self.assertEqual([{"field_one": [ListTuple("I", 0)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("I", 0)]}))
        # {"field_one": ["I", "am", "a", "list"]}
        self.assertEqual([{"field_one": [ListTuple("I", 0)]},
                          {"field_one": [ListTuple("I", 0), ListTuple("am", 1)]},
                          {"field_one": [ListTuple("I", 0), ListTuple("am", 1), ListTuple("a", 2)]},
                          {"field_one": [ListTuple("I", 0), ListTuple("am", 1), ListTuple("a", 2), ListTuple("list", 3)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("I", 0), ListTuple("am", 1), ListTuple("a", 2), ListTuple("list", 3)]}))

    def test_explode_structure_for_simple_dict(self):
        # {"field_one": "value_1", "field_two": "value_2"}
        self.assertEqual([{"field_one": "value_1"}, {"field_one": "value_1", "field_two": "value_2"}],
                         ExplodeStructure.explode_structure({"field_one": "value_1", "field_two": "value_2"}))

    def test_anonymous_string(self):
        # "string"
        self.assertEqual(["string"], ExplodeStructure.explode_structure("string"))

    def test_anonymous_bool(self):
        # True
        self.assertEqual([True], ExplodeStructure.explode_structure(True))

    def test_anonymous_int(self):
        # 1
        self.assertEqual([1], ExplodeStructure.explode_structure(1))

    def test_anonymous_list(self):
        # [1, 2]
        self.assertEqual([[ListTuple(1, 0)], [ListTuple(1, 0), ListTuple(2, 1)]], ExplodeStructure.explode_structure([1, 2]))

    def test_explode_structure_for_nested_list(self):
        # {"field_one": ["item_1", ["item_2"]]}
        self.assertEqual([{"field_one": [ListTuple("item_1", 0)]},
                          {"field_one": [ListTuple("item_1", 0), ListTuple([ListTuple("item_2", 0)], 1)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("item_1", 0),
                                                                                         ListTuple([ListTuple("item_2", 0)], 1)]}))

    def test_explode_structure_for_nested_list_with_multiple_items(self):
        # {"field_one": ["item_1", ["item_2", "item_3"]]}
        self.assertEqual([{"field_one": [ListTuple("item_1", 0)]},
                          {"field_one": [ListTuple("item_1", 0), ListTuple([ListTuple("item_2", 0)], 1)]},
                          {"field_one": [ListTuple("item_1", 0), ListTuple([ListTuple("item_2", 0),
                                                                            ListTuple("item_3", 1)], 1)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("item_1", 0),
                                                                                         ListTuple([ListTuple("item_2", 0),
                                                                                                    ListTuple("item_3", 1)], 1)]}))

    def test_explode_structure_for_nested_dicts(self):
        # {"field_one": {"nest_1": {"nest_2": "value_1", "nest_2_2": "value_2"}}}
        self.assertEqual([{"field_one": {"nest_1": {"nest_2": "value_1"}}},
                          {"field_one": {"nest_1": {"nest_2": "value_1", "nest_2_2": "value_2"}}}],
                         ExplodeStructure.explode_structure({"field_one": {"nest_1": {"nest_2": "value_1", "nest_2_2": "value_2"}}}))

    def test_explode_structure_preserves_list_item_indexing(self):
        # {"field_one": ["item_3", ["item_4"]]}
        self.assertEqual([{"field_one": [ListTuple("item_3", 2)]},
                          {"field_one": [ListTuple("item_3", 2), ListTuple("item_4", 3)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("item_3", 2),
                                                                                         ListTuple("item_4", 3)]}))

        self.assertEqual([{"field_one": [ListTuple("item_3", 2)]},
                          {"field_one": [ListTuple("item_3", 2), ListTuple([ListTuple("item_4", 0)], 3)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("item_3", 2),
                                                                                         ListTuple([ListTuple("item_4", 0)], 3)]}))

    def test_explode_structure_for_multi_nested_lists(self):
        # {"field_one": [50, "am", [[["a", "nested"]]], True]}
        self.assertEqual([{"field_one": [ListTuple(50, 0)]},
                          {"field_one": [ListTuple(50, 0), ListTuple("am", 1)]},
                          {"field_one": [ListTuple(50, 0), ListTuple("am", 1),
                                         ListTuple([ListTuple([ListTuple([ListTuple("a", 0)], 0)], 0)], 2)]},
                          {"field_one": [ListTuple(50, 0), ListTuple("am", 1),
                                         ListTuple([ListTuple([ListTuple([ListTuple("a", 0), ListTuple("nested", 1)], 0)], 0)], 2)]},
                          {"field_one": [ListTuple(50, 0), ListTuple("am", 1),
                                         ListTuple([ListTuple([ListTuple([ListTuple("a", 0), ListTuple("nested", 1)], 0)], 0)], 2),
                                         ListTuple(True, 3)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple(50, 0),
                                                                                         ListTuple("am", 1),
                                                                                         ListTuple([ListTuple([ListTuple([ListTuple("a", 0),
                                                                                                                          ListTuple("nested", 1)], 0)], 0)], 2),
                                                                                         ListTuple(True, 3)]}))

    def test_explode_structure_for_list_of_multi_nested_dicts(self):
        item_2_dict = {"item_2": {"nest_field_1": "nest_item_1_value"}}
        item_3_dict = {"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}
        item_4_dict = {"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                  "nest_field_1_key_2": "nest_item_2"},
                       "item_4_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}

        input_dict = {"list_one": [ListTuple("item_0", 0), ListTuple("item_1", 1), ListTuple(item_2_dict, 2), ListTuple(item_3_dict, 3), ListTuple(item_4_dict, 4)]}

        self.assertEqual([{"list_one": [ListTuple("item_0", 0)]},
                          {"list_one": [ListTuple("item_0", 0), ListTuple("item_1", 1)]},
                          {"list_one": [ListTuple("item_0", 0), ListTuple("item_1", 1), ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2)]},
                          {"list_one": [ListTuple("item_0", 0), ListTuple("item_1", 1), ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2),
                                        ListTuple({"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"}}}}}}, 3)]},
                          {"list_one": [ListTuple("item_0", 0), ListTuple("item_1", 1), ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2),
                                        ListTuple({"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                                 "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}, 3)]},
                          {"list_one": [ListTuple("item_0", 0),
                                        ListTuple("item_1", 1),
                                        ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2),
                                        ListTuple({"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                                 "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}, 3),
                                        ListTuple({"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"}}}, 4)]},
                          {"list_one": [ListTuple("item_0", 0),
                                        ListTuple("item_1", 1),
                                        ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2),
                                        ListTuple({"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                                 "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}, 3),
                                        ListTuple({"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                                              "nest_field_1_key_2": "nest_item_2"}}, 4)]},
                          {"list_one": [ListTuple("item_0", 0),
                                        ListTuple("item_1", 1),
                                        ListTuple({"item_2": {"nest_field_1": "nest_item_1_value"}}, 2),
                                        ListTuple({"item_3": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                                 "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}, 3),
                                        ListTuple({"item_4": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                                              "nest_field_1_key_2": "nest_item_2"},
                                                   "item_4_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}, 4)]}],
                         ExplodeStructure.explode_structure(input_dict))

    def test_explode_structure_handles_dict_of_nested_dicts(self):
        dict_1 = {"field_1": {"nest_field_1": "nest_item_1_value"}}
        dict_2 = {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                 "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}
        dict_3 = {"field_3": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                              "nest_field_1_key_2": "nest_item_2"},
                  "field_3_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}

        input_dict = {"dict_1": dict_1, "dict_2": dict_2, "dict_3": dict_3}
        self.assertEqual([{"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}}},
                          {"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}},
                           "dict_2": {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"}}}}}}},
                          {"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}},
                           "dict_2": {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}}},
                          {"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}},
                           "dict_2": {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}},
                           "dict_3": {"field_3": {"nest_field_1": {"nest_field_2": "nest_field_2_value"}}}},
                          {"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}},
                           "dict_2": {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}},
                           "dict_3": {"field_3": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                                  "nest_field_1_key_2": "nest_item_2"}}},
                          {"dict_1": {"field_1": {"nest_field_1": "nest_item_1_value"}},
                           "dict_2": {"field_2": {"nest_field_1": {"nest_field_2": {"nest_field_3": {"nest_field_4": {"nest_field_5": "nest_field_5_value"},
                                                                                                     "nest_field_4_key_2": {"nest_field_5_key_2": "nest_field_5_value_2"}}}}}},
                           "dict_3": {"field_3": {"nest_field_1": {"nest_field_2": "nest_field_2_value"},
                                                  "nest_field_1_key_2": "nest_item_2"},
                                      "field_3_key_2": {"nest_field_3_key_2": "nest_field_3_key_2_value_2"}}}],
                         ExplodeStructure.explode_structure(input_dict))

    def test_explode_structure_for_list_of_nested_lists_and_dicts(self):
        # {"field_one": ["A", [["nested", "list"]], "item"], {"lvl1": {"lvl2": {"item_1": "nested",
                                                                              # "item_2": [["item_2_nested"]]}}}]}
        self.assertEqual([{"field_one": [ListTuple("A", 0)]},
                          {"field_one": [ListTuple("A", 0), ListTuple([ListTuple([ListTuple([ListTuple("nested", 0)], 0)], 0)], 1)]},
                          {"field_one": [ListTuple("A", 0),
                                         ListTuple([ListTuple([ListTuple([ListTuple("nested", 0), ListTuple("list", 1)], 0)], 0)], 1)]},
                          {"field_one": [ListTuple("A", 0),
                                         ListTuple([ListTuple([ListTuple([ListTuple("nested", 0), ListTuple("list", 1)], 0)], 0), ListTuple("item", 1)], 1)]},
                          {"field_one": [ListTuple("A", 0),
                                         ListTuple([ListTuple([ListTuple([ListTuple("nested", 0), ListTuple("list", 1)], 0)], 0), ListTuple("item", 1)], 1),
                                         ListTuple({"lvl1": {"lvl2": {"item_1": "nested"}}}, 2)]},
                          {"field_one": [ListTuple("A", 0),
                                         ListTuple([ListTuple([ListTuple([ListTuple("nested", 0), ListTuple("list", 1)], 0)], 0), ListTuple("item", 1)], 1),
                                         ListTuple({"lvl1": {"lvl2": {"item_1": "nested", "item_2": ListTuple([ListTuple([ListTuple("item_2_nested", 0)], 0)], 0)}}}, 2)]}],
                         ExplodeStructure.explode_structure({"field_one": [ListTuple("A", 0),
                                                                                         ListTuple([ListTuple([ListTuple([ListTuple("nested", 0),
                                                                                                                          ListTuple("list", 1)], 0)], 0),
                                                                                                    ListTuple("item", 1)], 1),
                                                                                         ListTuple({"lvl1": {"lvl2": {"item_1": "nested", "item_2": ListTuple([ListTuple([ListTuple("item_2_nested", 0)], 0)], 0)}}}, 2)]}))

    # def test_explode_structure_handles_wide_and_deep_nested_mixtures_of_lists_and_dicts(self):
    #     nested_dict = {"lvl1": {"lvl2_1": {"lvl3": "value_6"},
    #                             "lvl2_2": "value_7",
    #                             "lvl2_3": [["value_8"],
    #                                        {"lvl3": {"lvl4": "value_9"}},
    #                                        {"lvl3": "value_10"}]},
    #                    "lvl1_2": {"lvl2_1": "value_11"},
    #                    "lvl1_3": {"lvl2_1": {"lvl3_1": {"lvl4_1": "value_12"}}}}
    #     main_dict = {"nested_dict": nested_dict}
    #
    #     self.assertEqual([{"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7"}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7", "lvl2_3": [ListTuple([ListTuple("value_8", 0)], 0)]}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7", "lvl2_3": [ListTuple([ListTuple("value_8", 0)], 0), ListTuple({"lvl3": {"lvl4": "value_9"}}, 1)]}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7", "lvl2_3": [ListTuple([ListTuple("value_8", 0)], 0), ListTuple({"lvl3": {"lvl4": "value_9"}}, 1), ListTuple({"lvl3": "value_10"}, 2)]}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7", "lvl2_3": [ListTuple([ListTuple("value_8", 0)], 0), ListTuple({"lvl3": {"lvl4": "value_9"}}, 1), ListTuple({"lvl3": "value_10"}, 2)]}, "lvl1_2": {"lvl2_1": "value_11"}}},
    #                       {"nested_dict": {"lvl1": {"lvl2_1": {"lvl3": "value_6"}, "lvl2_2": "value_7", "lvl2_3": [ListTuple([ListTuple("value_8", 0)], 0), ListTuple({"lvl3": {"lvl4": "value_9"}}, 1), ListTuple({"lvl3": "value_10"}, 2)]}, "lvl1_2": {"lvl2_1": "value_11"}, "lvl1_3": {"lvl2_1": {"lvl3_1": {"lvl4_1": "value_12"}}}}}],
    #                      ExplodeStructure.explode_structure(main_dict))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

from .explode_structure import ExplodeStructure


class RemoveData:

    @classmethod
    def subtract_data_from_input(cls, current_input, sisl_dict_elements_to_delete):
        leaves_to_delete = ExplodeStructure.enumerate_paths(sisl_dict_elements_to_delete)
        return cls.remove_matching_elements(current_input, leaves_to_delete)

    @classmethod
    def remove_matching_elements(cls, current_input, leaves_to_delete):
        for each_leaf in leaves_to_delete:
            current_input = cls.do_remove(current_input, each_leaf)

        return current_input

    @classmethod
    def do_remove(cls, current_input, each_leaf):
        if type(each_leaf) == dict:
            cls.handle_each_dict(current_input, each_leaf)
        elif type(each_leaf) == list:
            cls.handle_each_list(current_input, each_leaf)
        else:
            if current_input == each_leaf:
                current_input = {}

        return current_input

    @classmethod
    def handle_each_dict(cls, current_input, each_leaf):
        for key in each_leaf:
            if type(each_leaf[key]) == dict:
                cls.do_remove(current_input[key], each_leaf[key])
                if current_input[key] == {}:
                    del current_input[key]
            if type(each_leaf[key]) == list:
                if type(each_leaf[key][0].value) == dict:
                    cls.do_remove(current_input[key][0].value, each_leaf[key][0].value)
                    if current_input[key][0].value == {}:
                        current_input[key].pop(0)

                if type(each_leaf[key][0].value) == list:
                    cls.do_remove(current_input[key][0].value, each_leaf[key][0].value)

                    if not current_input[key][0].value:
                        current_input[key].pop(0)

                if not current_input[key]:
                    del current_input[key]

            if key in current_input and each_leaf[key] == current_input[key]:
                del current_input[key]

            elif (key in current_input) and (type(each_leaf[key]) == list) and (each_leaf[key][0] in current_input[key]):
                current_input[key].pop(0)

    @classmethod
    def handle_each_list(cls, current_input, each_leaf):
        for index, value in enumerate(each_leaf):
            if type(each_leaf[index].value) == dict:
                cls.do_remove(current_input[index].value, each_leaf[index].value)
                if current_input[index].value == {}:
                    del current_input[index]
            if type(each_leaf[index].value) == list:
                cls.do_remove(current_input[index].value, each_leaf[index].value)
                if not current_input[index].value:
                    current_input.pop(0)
            if each_leaf[index] in current_input:
                current_input.pop(0)



# Copyright PA Knowledge Ltd 2021

import unittest
from pysisl.sisl_decoder import SislDecoder
from pysisl.parser_error import ParserError


class SislRawTypesParsingTests(unittest.TestCase):
    def test_parser_error(self):
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, "{")
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, "{1abc: !type \"value\"}")
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, "{abc: type \"value\"}")
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, "{abc: !type \"value}")

    def test_basic_sisl_structures(self):
        self.assertEqual({}, SislDecoder().parse_raw_types("{}"))
        self.assertEqual({"abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{abc: !type \"value\"}"))
        self.assertEqual({
            "abc": {
                "type": "type",
                "value": {
                    "nested": {
                        "type": "type",
                        "value": "value"
                    }
                }
            }
        }, SislDecoder().parse_raw_types("{abc: !type {nested: !type \"value\"}}"))

        self.assertEqual({
            "abc": {
                "type": "type",
                "value": {
                    "nested": {
                        "type": "type",
                        "value": {
                            "nested": {
                                "type": "type",
                                "value": "value"
                            }
                        }
                    }
                }
            }
        }, SislDecoder().parse_raw_types("{abc: !type {nested: !type {nested: !type \"value\"}}}"))

    def test_basic_values(self):
        self.assertEqual({"abc": {"type": "type", "value": "value1"}}, SislDecoder().parse_raw_types("{abc: !type \"value1\"}"))
        self.assertEqual({"abc": {"type": "type", "value": "VALUE1"}}, SislDecoder().parse_raw_types("{abc: !type \"VALUE1\"}"))
        self.assertEqual({"abc": {"type": "type", "value": r'val\"ue'}}, SislDecoder().parse_raw_types(r'{abc: !type "val\"ue"}'))
        self.assertEqual({"abc": {"type": "type", "value": r'val\\ue'}}, SislDecoder().parse_raw_types(r'{abc: !type "val\\ue"}'))
        self.assertEqual({"abc": {"type": "type", "value": r'val ue'}}, SislDecoder().parse_raw_types(r'{abc: !type "val ue"}'))

    def test_basic_names(self):
        self.assertEqual({"aBc": {"type": "type", "value": "value1"}}, SislDecoder().parse_raw_types("{aBc: !type \"value1\"}"))
        self.assertEqual({"AbC": {"type": "type", "value": "value1"}}, SislDecoder().parse_raw_types("{AbC: !type \"value1\"}"))
        self.assertEqual({"_B.-": {"type": "type", "value": "value1"}}, SislDecoder().parse_raw_types("{_B.-: !type \"value1\"}"))

    def test_basic_types(self):
        self.assertEqual({"abc": {"type": "aBc", "value": "value1"}}, SislDecoder().parse_raw_types("{abc: !aBc \"value1\"}"))
        self.assertEqual({"abc": {"type": "AbC", "value": "value1"}}, SislDecoder().parse_raw_types("{abc: !AbC \"value1\"}"))
        self.assertEqual({"abc": {"type": "_B.-", "value": "value1"}}, SislDecoder().parse_raw_types("{abc: !_B.- \"value1\"}"))

        self.assertEqual({"abc": {"type": f"str{'r' * 252}", "value": "test"}}, SislDecoder().parse_raw_types(f'{{abc: !str{"r" * 252} \"test\"}}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !str{"r" * 253} \"test\"}}')

    def test_lists(self):
        self.assertEqual({"abc": {"type": "type", "value": "value"}, "def": {"type": "typeb", "value": "value2"}},
                         SislDecoder().parse_raw_types("{abc: !type \"value\",def: !typeb \"value2\"}"))

        self.assertEqual({"abc": {"type": "type", "value": "value"},
                          "def": {"type": "typeb", "value": "value2"},
                          "ghi": {"type": "typec", "value": "value3"}},
                         SislDecoder().parse_raw_types("{abc: !type \"value\","
                                                       "def: !typeb \"value2\","
                                                       "ghi: !typec \"value3\"}"))

    def test_lists_with_whitespace_after_separator(self):
        self.assertEqual({
            "abc": {"type": "type", "value": "value"},
            "def": {"type": "typeb", "value": "value2"}
        }, SislDecoder().parse_raw_types("{abc: !type \"value\", def: !typeb \"value2\"}"))
        self.assertEqual({"abc": {"type": "type", "value": "value"}, "def": {"type": "typeb", "value": "value2"}},
                         SislDecoder().parse_raw_types("{abc: !type \"value\",  def: !typeb \"value2\"}"))

    def test_lists_with_whitespace_before_separator(self):
        self.assertEqual({"abc": {"type": "type", "value": "value"}, "def": {"type": "typeb", "value": "value2"}},
                         SislDecoder().parse_raw_types("{abc: !type \"value\" , def: !typeb \"value2\"}"))

        self.assertEqual({"abc": {"type": "type", "value": "value"}, "def": {"type": "typeb", "value": "value2"}},
                         SislDecoder().parse_raw_types("{abc: !type \"value\" , def: !typeb \"value2\"}"))

    def test_whitespace_limit(self):
        self.assertEqual({"abc": {"type": "type", "value": "test"}}, SislDecoder().parse_raw_types(f'{{{" " * 255}abc: !type \"test\"}}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{{" " * 256}abc: !type \"test\"}}')

        self.assertEqual({"abc": {"type": "type", "value": "test"}}, SislDecoder().parse_raw_types(f'{{abc:{" " * 255}!type \"test\"}}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, r'{abc:!type \"test\"}')
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc:{" " * 256}!type \"test\"}}')

        self.assertEqual({"abc": {"type": "type", "value": "test"}}, SislDecoder().parse_raw_types(f'{{abc: !type{" " * 255}\"test\"}}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !type\"test\"}}')
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !type{" " * 300}\"test\"}}')

        self.assertEqual({"abc": {"type": "type", "value": "test1"}, "def": {"type": "type", "value": "test2"}}, SislDecoder().parse_raw_types(f'{{abc: !type \"test1\"{" " * 255}, def: !type \"test2\"{" " * 255}}}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !type \"test\"{" " * 256}}}')
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !type \"test1\"{" " * 256}, def: !type \"test2\"}}')

        self.assertEqual({"abc": {"type": "type", "value": "test"}}, SislDecoder().parse_raw_types(f'{{abc: !type \"test\"}}{" " * 255}'))
        self.assertRaises(ParserError, SislDecoder().parse_raw_types, f'{{abc: !type \"test\"}}{" " * 256}')

    def test_nested_lists(self):
        self.assertEqual({'field_one': {'type': 'list',
                                        'value': {'_0': {'type': 'int', 'value': '1'},
                                                  '_1': {'type': 'list',
                                                         'value': {'_0': {'type': 'int',
                                                                          'value': '2'}}}}}},
                         SislDecoder().parse_raw_types('{field_one: !list {_0: !int "1", _1: !list {_0: !int "2"}}}'))

    def test_name_special_chars(self):
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{_abc: !type \"value\"}"))
        self.assertEqual({"ab_c_": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{ab_c_: !type \"value\"}"))
        self.assertEqual({"ab.c": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{ab.c: !type \"value\"}"))
        self.assertEqual({"ab3c": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{ab3c: !type \"value\"}"))

    def test_type_special_chars(self):
        self.assertEqual({"abc": {"type": "_abc", "value": "value"}}, SislDecoder().parse_raw_types("{abc: !_abc \"value\"}"))
        self.assertEqual({"abc": {"type": "ab_c", "value": "value"}}, SislDecoder().parse_raw_types("{abc: !ab_c \"value\"}"))
        self.assertEqual({"abc": {"type": "ab.c", "value": "value"}}, SislDecoder().parse_raw_types("{abc: !ab.c \"value\"}"))
        self.assertEqual({"abc": {"type": "ab3c", "value": "value"}}, SislDecoder().parse_raw_types("{abc: !ab3c \"value\"}"))
        self.assertEqual({"abc": {"type": "t", "value": "value1"}}, SislDecoder().parse_raw_types("{abc: !t \"value1\"}"))
        self.assertEqual({"c": {"type": "t", "value": "value1"}}, SislDecoder().parse_raw_types("{c: !t \"value1\"}"))

    def test_value_special_chars(self):
        self.assertEqual({"_abc": {"type": "type", "value": "value="}}, SislDecoder().parse_raw_types("{_abc: !type \"value=\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value-"}}, SislDecoder().parse_raw_types("{_abc: !type \"value-\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value:"}}, SislDecoder().parse_raw_types("{_abc: !type \"value:\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value."}}, SislDecoder().parse_raw_types("{_abc: !type \"value.\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value+"}}, SislDecoder().parse_raw_types("{_abc: !type \"value+\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value@"}}, SislDecoder().parse_raw_types("{_abc: !type \"value@\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value'"}}, SislDecoder().parse_raw_types("{_abc: !type \"value'\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value!"}}, SislDecoder().parse_raw_types("{_abc: !type \"value!\"}"))

    def test_value_unicode_chars(self):
        self.assertEqual({"_abc": {"type": "type", "value": "ӧӧФ@፨="}}, SislDecoder().parse_raw_types("{_abc: !type \"ӧӧФ@፨=\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "ӧ"}}, SislDecoder().parse_raw_types("{_abc: !type \"\u04E7\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "\u04E7"}}, SislDecoder().parse_raw_types("{_abc: !type \"\u04E7\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "U+274C"}}, SislDecoder().parse_raw_types("{_abc: !type \"U+274C\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "&#10060;"}}, SislDecoder().parse_raw_types("{_abc: !type \"&#10060;\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "£¥"}}, SislDecoder().parse_raw_types("{_abc: !type \"£¥\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "😂❤🔥❌"}}, SislDecoder().parse_raw_types("{_abc: !type \"😂❤🔥❌\"}"))

    def test_allow_space(self):
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{_abc: !type \"value\"} "))
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{ _abc: !type \"value\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{_abc: !type \"value\" }"))
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{_abc: !type  \"value\"}"))
        self.assertEqual({"_abc": {"type": "type", "value": "value"}}, SislDecoder().parse_raw_types("{_abc:  !type \"value\"}"))

    def test_allow_escaped_quote_followed_by_space(self):
        self.assertEqual({"_abc": {"type": "type", "value": r'value\" '}}, SislDecoder().parse_raw_types(r'{_abc: !type "value\" " }'))

    def test_allow_escaped_quote_followed_by_commas(self):
        self.assertEqual({"_abc": {"type": "type", "value": r'\",'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\","}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\",\"'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\",\""}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\",\",'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\",\","}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\",KB'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\",KB"}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\\,KB'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\\,KB"}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\\\",KB'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\\\",KB"}'))

    def test_allow_escaped_characters_before_final_quotes(self):
        self.assertEqual({"_abc": {"type": "type", "value": r'\\'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\\"}'))
        self.assertEqual({"_abc": {"type": "type", "value": r'\\\"'}}, SislDecoder().parse_raw_types(r'{_abc: !type "\\\""}'))

    def test_example_syslog_sisl(self):
        self.assertEqual({'app-name': {'type': 'string', 'value': 'containerd'},
                          'date': {'type': 'string', 'value': '2019-08-08T13:50:14.170225+01:00'},
                          'host': {'type': 'string', 'value': 'centosvm'},
                          'message': {'type': 'string',
                                      'value': '  time=2020-12-09T16:47:16.358818934Z'}},
                         SislDecoder().parse_raw_types(r'{date: !string "2019-08-08T13:50:14.170225+01:00", host: !string "centosvm", app-name: !string "containerd", message: !string "  time=2020-12-09T16:47:16.358818934Z"}'))


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

from pysisl.deep_merger import ListTuple


class Anon:
    def __init__(self, value):
        self.value = value


class SislEncoder:

    @classmethod
    def dumps(cls, decoded_sisl):
        return "{" + cls._concatenate_sisl_triple(decoded_sisl) + "}"

    @classmethod
    def _concatenate_sisl_triple(cls, decoded_sisl):
        return ', '.join(cls._sisl_triple_list(cls.construct_dict_for_anonymous_data(decoded_sisl)))

    @classmethod
    def construct_dict_for_anonymous_data(cls, decoded_sisl):
        if type(decoded_sisl) != dict:
            return {"_": Anon(decoded_sisl)}
        else:
            return decoded_sisl

    @classmethod
    def _sisl_triple_list(cls, dict_to_encode):
        return [cls._construct_sisl_triple(k, v) for k, v in dict_to_encode.items()]

    @classmethod
    def _construct_sisl_triple(cls, key, value):
        return f"{key}: {cls._select_type_and_value(value)}"

    @classmethod
    def _select_type_and_value(cls, value):
        if type(value) == Anon:
            return f"!_{cls.sisl_string_from_type(value.value)}"
        else:
            return f"!{cls.sisl_string_from_type(value)}"

    @classmethod
    def sisl_string_from_type(cls, value):
        try:
            return {
                int: lambda val: f"int \"{value}\"",
                float: lambda val: f"float \"{value}\"",
                str: lambda val: f"str \"{cls._escape(value)}\"",
                bool: lambda val: f"bool \"{str(value).lower()}\"",
                list: lambda val: f"list {{{cls._construct_sisl_for_list_object(value)}}}",
                dict: lambda val: f"obj {cls.dumps(value)}",
                None.__class__: lambda val: f"null \"\""
            }[type(value)](value)
        except KeyError as err:
            raise TypeValidationError(err)

    @classmethod
    def _escape(cls, value: str):
        return value.encode('unicode_escape').decode().replace(r'"', r'\"')

    @classmethod
    def _construct_sisl_for_list_object(cls, value):
        return ', '.join([cls._construct_sisl_triple(*cls._check_if_list_tuple(i, item)) for i, item in enumerate(value)])

    @classmethod
    def _check_if_list_tuple(cls, index, list_item):
        if type(list_item) is ListTuple:
            return f"_{list_item.position}", list_item.value

        return f"_{index}", list_item


class TypeValidationError(Exception):
    pass



# Copyright PA Knowledge Ltd 2021

from .parser_error import ParserError


def p_sisl_object(p):
    """sisl_object : BEGIN_OBJECT sisl_triplets END_OBJECT """
    p[0] = p[2]


def p_sisl_triplets(p):
    """sisl_triplets : empty_triplet
                     | sisl_triplet
                     | sisl_triplet_list """
    p[0] = p[1]


def p_empty_triplet(p):
    """ empty_triplet : """
    p[0] = {}


def p_sisl_triplet(p):
    """ sisl_triplet : NAME type_and_value """
    p[0] = {p[1]: p[2]}


def p_type_and_value(p):
    """ type_and_value : type value """
    p[0] = dict(p[1], **p[2])


def p_type(p):
    """ type : TYPE """
    p[0] = {"type": p[1]}


def p_value(p):
    """ value : VALUE
              | sisl_object """
    p[0] = {"value": p[1]}


def p_sisl_triplet_list(p):
    """ sisl_triplet_list : sisl_triplet LIST_SEPARATOR sisl_triplets """
    p[0] = p[3]
    p[0].update(p[1])


def p_error(p):
    raise ParserError(p)



import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class PostDevelopCommand(develop):
    def run(self):
        develop.run(self)
        import pysisl
        pysisl.loads("{}")


class PostInstallCommand(install):
    def run(self):
        install.run(self)
        import pysisl
        pysisl.loads("{}")


setuptools.setup(
    name="pysisl",
    version="0.0.13",
    author="Oakdoor",
    author_email="info@oakdoor.io",
    description="A python library for serialising and deserialising SISL (Simple Information Serialization Language)",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://oakdoor.io",
    project_urls={
        'Project': 'https://oakdoor.io',
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    setup_requires=['ply', 'jsonschema', 'deepmerge', 'mergedeep', 'construct'],
    install_requires=['ply', 'jsonschema', 'deepmerge', 'mergedeep', 'construct'],
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
    python_requires='>=3.6'
)



# Copyright PA Knowledge Ltd 2021
import unittest
from pysisl import sisl_wrapper
from binascii import unhexlify


class SislWrapperTests(unittest.TestCase):

    def test_simple_unwrapping_with_key(self):
        wrapped_data = b"\xd1\xdf\x5f\xff\x00\x01\x00\x00\x00\x00\x00\x30\x00\x00\x00\x01" \
                       b"\x00\x03\x00\x08\x80\x7f\xd4\x1f\x1b\x51\x27\xa9\x00\x00\x00\x00" \
                       b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x5f\xdf\xd1" \
                       b"\xe8\x1a\xb8\x73\x74"
        self.assertEqual(b'hello', sisl_wrapper.SislWrapper().unwraps(wrapped_data))

    def test_wrapping_byte_string_then_unwrap_returns_original_data(self):
        data = b'hello'
        wrapped_data = sisl_wrapper.SislWrapper().wraps(data)
        unwrapped_data = sisl_wrapper.SislWrapper().unwraps(wrapped_data)
        self.assertEqual(data, unwrapped_data)

    def test_wrapping_bytearray_then_unwrap_returns_original_data(self):
        data = bytearray('hello', encoding='utf-8')
        wrapped_data = sisl_wrapper.SislWrapper().wraps(data)
        unwrapped_data = sisl_wrapper.SislWrapper().unwraps(wrapped_data)
        self.assertEqual(data, unwrapped_data)

    def test_wrapped_payload_includes_cloaked_dagger_header(self):
        data = b'hello'
        wrapped_data = sisl_wrapper.SislWrapper().wraps(data)
        sisl_wrapper.CloakDagger.check_valid_cloaked_dagger_header(wrapped_data)

    def test_empty_payload_returns_only_cloaked_dagger_header(self):
        data = b''
        wrapped_data = sisl_wrapper.SislWrapper().wraps(data)
        self.assertEqual(self.expected_wrapped_data_size(data), len(wrapped_data))
        sisl_wrapper.CloakDagger.check_valid_cloaked_dagger_header(wrapped_data)

    def test_one_char_payload_returns_wrapped_string(self):
        data = b'a'
        wrapped_data = sisl_wrapper.SislWrapper().wraps(data)
        self.assertEqual(self.expected_wrapped_data_size(data), len(wrapped_data))
        sisl_wrapper.CloakDagger.check_valid_cloaked_dagger_header(wrapped_data)

    def test_throws_error_when_unwrapping_not_wrapped_payload(self):
        self.assertRaises(sisl_wrapper.InvalidHeaderError, sisl_wrapper.SislWrapper().unwraps, b'abcdef')

    def test_throws_error_when_unwrapping_not_wrapped_empty_payload(self):
        self.assertRaises(sisl_wrapper.InvalidHeaderError, sisl_wrapper.SislWrapper().unwraps, b'')

    def test_throw_error_when_wrapping_payload_of_invalid_type(self):
        self.assertRaises(TypeError, sisl_wrapper.SislWrapper().wraps, {1: 3})

    def test_throws_error_when_unwrapping_empty_payload_of_invalid_type(self):
        self.assertRaises(TypeError, sisl_wrapper.SislWrapper().unwraps, '')

    def test_throws_error_when_unwrapping_file_with_invalid_header(self):
        wrapped_data = b"\xd1\xdf\x5f\xff\x00\x01\x00\x00\x00\x00\x00\x30\x00\x00\x00\x01" \
                       b"\x00\x03\x00\x08\x80\x7f\xd4\x1f\x1b\x51\x27\xa9\x00\x00\x00\x00" \
                       b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1f\x5f\xdf\xd1" \
                       b"\xe8\x1a\xb8\x73\x74"
        self.assertRaises(sisl_wrapper.InvalidHeaderError, sisl_wrapper.SislWrapper().unwraps, wrapped_data)

    def test_key_generator_does_not_include_0(self):
        for i in range(10):
            self.assertTrue(b'\x00' not in sisl_wrapper.SislWrapper.key_generation())

    def test_unwrap_file_stream(self):
        wrapped_data = b"\xd1\xdf\x5f\xff\x00\x01\x00\x00\x00\x00\x00\x30\x00\x00\x00\x01" \
                       b"\x00\x03\x00\x08\x80\x7f\xd4\x1f\x1b\x51\x27\xa9\x00\x00\x00\x00" \
                       b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x5f\xdf\xd1" \
                       b"\xe8\x1a\xb8\x73\x74"
        unwrapped_data = b""
        wrapper = sisl_wrapper.SislWrapper()
        while len(wrapped_data) > 0:
            chunk, wrapped_data = SislWrapperTests.read_chunk(wrapped_data)
            unwrapped_data += wrapper.unwraps(chunk)
        self.assertEqual(b"hello", unwrapped_data)

    def test_unwrap_file_stream_raises_exception_if_chunk_smaller_than_cd_header(self):
        wrapped_data = b"\xd1\xdf\x5f\xff"
        wrapper = sisl_wrapper.SislWrapper()
        self.assertRaises(Exception, wrapper.unwraps, wrapped_data)

    def test_unwrap_file_stream_raises_exception_if_file_not_wrapped(self):
        data = b"\xe8\x1a\xb8\x73\x74\xe8\x1a\xb8\x73\x74\xe8\x1a" \
               b"\xe8\x1a\xb8\x73\x74\xe8\x1a\xb8\x73\x74\xe8\x1a" \
               b"\xe8\x1a\xb8\x73\x74\xe8\x1a\xb8\x73\x74\xe8\x1a" \
               b"\xe8\x1a\xb8\x73\x74\xe8\x1a\xb8\x73\x74\xe8\x1a"
        wrapper = sisl_wrapper.SislWrapper()
        self.assertRaises(Exception, wrapper.unwraps, data)

    @staticmethod
    def read_chunk(data):
        return data[:48], data[48:]

    @staticmethod
    def fake_key():
        return unhexlify(b"0123456789ABCDEF")

    @staticmethod
    def expected_wrapped_data_size(data):
        return sisl_wrapper.SislWrapper().headerLen + len(data)


if __name__ == '__main__':
    unittest.main()



# Copyright PA Knowledge Ltd 2021

from mergedeep import merge, Strategy


class JoinSisl:

    @classmethod
    def join(cls, list_of_dicts):
        return merge({}, *list_of_dicts, strategy=Strategy.TYPESAFE_ADDITIVE)
